#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from unittest import mock

from parameterized import parameterized

from evo.common import HTTPHeaderDict
from evo.common.test_tools import MockResponse, TestHTTPHeaderDict, long_test
from evo.oauth import AccessToken, AnyScopes, AuthorizationCodeAuthorizer, EvoScopes, OAuthError

from ._helpers import (
    AUTHORIZATION_CODE,
    CLIENT_ID,
    CLIENT_SECRET,
    REDIRECT_URL,
    STATE_TOKEN,
    VERIFIER_TOKEN,
    TestWithOAuthConnector,
    get_access_token,
    patch_urlsafe_tokens,
    patch_webbrowser_open,
)

AUTH_HOSTNAME = ...


class TestAuthorizationCodeAuthorizer(TestWithOAuthConnector):
    def setUp(self) -> None:
        super().setUp()
        self.authorizer = AuthorizationCodeAuthorizer(oauth_connector=self.connector, redirect_url=REDIRECT_URL)
        self.first_token = get_access_token(access_token="first_token")
        self.second_token = get_access_token(access_token="second_token")
        self.transport.request.side_effect = [
            MockResponse(status_code=200, content=self.first_token.model_dump_json(by_alias=True, exclude_unset=True)),
            MockResponse(
                status_code=200,
                content=self.second_token.model_dump_json(by_alias=True, exclude_unset=True),
            ),
        ]

    async def _login(self, scopes: AnyScopes | None = None, timeout_seconds: int | None = None) -> None:
        if scopes is not None:
            self.authorizer._scopes = scopes

        if timeout_seconds is not None:
            await self.authorizer.login(timeout_seconds=timeout_seconds)
        else:
            await self.authorizer.login()

    async def login(
        self,
        scopes: AnyScopes | None = None,
        timeout_seconds: int | None = None,
        authenticate: bool = True,
        state: str = STATE_TOKEN,
        verifier: str = VERIFIER_TOKEN,
    ) -> mock.Mock:
        """Log in and return the mock object that was used to patch the `webbrowser.open` function.

        :param scopes: The OAuth scopes to use in this test. If `None`, the default scopes are used.
        :param timeout_seconds: The timeout to use for the login. If `None`, the timeout parameter is not passed to the
            `login` method.
        :param authenticate: Whether to authenticate the user during the login.
        :param state: The state token to use for the login.
        :param verifier: The verifier token to use for the login.
        """
        if authenticate:
            with (
                patch_webbrowser_open(authenticate=True) as mock_open,
                patch_urlsafe_tokens(state=state, verifier=verifier),
            ):
                await self._login(scopes, timeout_seconds)
        else:
            with (
                patch_webbrowser_open(authenticate=False) as mock_open,
                patch_urlsafe_tokens(state=state, verifier=verifier),
                self.assertRaises(OAuthError),
            ):
                await self._login(scopes, timeout_seconds)

        return mock_open

    def assert_token_equals(self, expected_token: AccessToken | None) -> None:
        """Check that the access token is as expected."""
        actual_token: AccessToken | None = self.authorizer._BaseAuthorizer__token
        self.assertEqual(expected_token, actual_token)

    async def test_login_defaults(self) -> None:
        """Test the login method of the AuthorizationCodeAuthorizer."""
        mock_open = await self.login()
        expected_auth_url = self.get_expected_auth_url(
            state=STATE_TOKEN, verifier=VERIFIER_TOKEN, scopes=EvoScopes.default
        )
        mock_open.assert_called_once_with(expected_auth_url)
        self.assert_fetched_token(
            grant_type="authorization_code",
            code=AUTHORIZATION_CODE,
            code_verifier=VERIFIER_TOKEN,
            redirect_uri=REDIRECT_URL,
            client_id=CLIENT_ID,
            client_secret=CLIENT_SECRET,
        )
        self.assert_token_equals(self.first_token)

    async def test_login_custom_scopes(self) -> None:
        """Test the login method of the AuthorizationCodeAuthorizer with custom scopes."""
        mock_open = await self.login(scopes=EvoScopes.all_evo)
        expected_auth_url = self.get_expected_auth_url(
            state=STATE_TOKEN, verifier=VERIFIER_TOKEN, scopes=EvoScopes.all_evo
        )
        mock_open.assert_called_once_with(expected_auth_url)
        self.assert_fetched_token(
            grant_type="authorization_code",
            code=AUTHORIZATION_CODE,
            code_verifier=VERIFIER_TOKEN,
            redirect_uri=REDIRECT_URL,
            client_id=CLIENT_ID,
            client_secret=CLIENT_SECRET,
        )
        self.assert_token_equals(self.first_token)

    @parameterized.expand(
        [
            ("no delay", 0),
            ("short delay", 1),
            ("longer delay", 5),
        ]
    )
    @long_test
    async def test_login_custom_timeout(self, _label: str, delay: int) -> None:
        """Test the login method of the AuthorizationCodeAuthorizer with a custom timeout."""
        mock_open = await self.login(timeout_seconds=delay, authenticate=False)
        expected_auth_url = self.get_expected_auth_url(
            state=STATE_TOKEN, verifier=VERIFIER_TOKEN, scopes=EvoScopes.default
        )
        mock_open.assert_called_once_with(expected_auth_url)
        self.assert_token_equals(None)

    @parameterized.expand(
        [
            ("no delay", 0),
            ("short delay", 1),
            ("longer delay", 5),
        ]
    )
    @long_test
    async def test_login_custom_scopes_timeout(self, _label: str, delay: int) -> None:
        """Test the login method of the AuthorizationCodeAuthorizer with custom scopes and a custom timeout."""
        mock_open = await self.login(scopes=EvoScopes.all_evo, timeout_seconds=delay, authenticate=False)
        expected_auth_url = self.get_expected_auth_url(
            state=STATE_TOKEN, verifier=VERIFIER_TOKEN, scopes=EvoScopes.all_evo
        )
        mock_open.assert_called_once_with(expected_auth_url)
        self.assert_token_equals(None)

    async def test_login(self) -> None:
        """Test the login method of the AuthorizationCodeAuthorizer."""
        mock_open = await self.login()
        expected_auth_url = self.get_expected_auth_url(
            state=STATE_TOKEN, verifier=VERIFIER_TOKEN, scopes=EvoScopes.default
        )
        mock_open.assert_called_once_with(expected_auth_url)
        self.assert_fetched_token(
            grant_type="authorization_code",
            code=AUTHORIZATION_CODE,
            code_verifier=VERIFIER_TOKEN,
            redirect_uri=REDIRECT_URL,
            client_id=CLIENT_ID,
            client_secret=CLIENT_SECRET,
        )
        self.assert_token_equals(self.first_token)

    async def test_refresh_token(self) -> None:
        """Test the refresh_token method of the AuthorizationCodeAuthorizer."""
        await self.login()
        self.transport.request.reset_mock()
        self.assertTrue(await self.authorizer.refresh_token())
        self.assert_fetched_token(
            grant_type="refresh_token",
            refresh_token=self.first_token.refresh_token,
            client_id=CLIENT_ID,
            client_secret=CLIENT_SECRET,
        )
        self.assert_token_equals(self.second_token)

    async def test_assert_refresh_token_fails_before_login_fails(self) -> None:
        """Assert that refreshing the token fails before logging in."""
        with self.assertRaises(OAuthError):
            await self.authorizer.refresh_token()
        self.transport.assert_no_requests()

    async def test_refresh_token_fails_with_http_error(self) -> None:
        """Test that the refresh_token method of the AuthorizationCodeAuthorizer fails when an HTTP error occurs."""
        await self.login()
        self.assert_token_equals(self.first_token)
        self.transport.request.reset_mock()
        self.transport.request.side_effect = [MockResponse(status_code=500)]
        self.assertFalse(await self.authorizer.refresh_token())
        self.assert_fetched_token(
            grant_type="refresh_token",
            refresh_token=self.first_token.refresh_token,
            client_id=CLIENT_ID,
            client_secret=CLIENT_SECRET,
        )
        self.assert_token_equals(self.first_token)

    async def test_get_default_headers(self) -> None:
        """Test the get_default_headers method of the AuthorizationCodeAuthorizer."""
        await self.login()
        headers = await self.authorizer.get_default_headers()
        self.assertIs(type(headers), HTTPHeaderDict, "The default headers should be an HTTPHeaderDict.")
        self.assertEqual(
            TestHTTPHeaderDict({"Authorization": f"Bearer {self.first_token.access_token}"}),
            TestHTTPHeaderDict(headers),
        )

    async def test_get_default_headers_before_login_fails(self) -> None:
        """Test that the get_default_headers method of the AuthorizationCodeAuthorizer fails before logging in."""
        with self.assertRaises(OAuthError):
            await self.authorizer.get_default_headers()

    async def test_refresh_token_fails_without_refresh_token(self) -> None:
        """Test refreshing an access token fails if the current token does not have a refresh token"""
        with self.assertRaises(OAuthError):
            await self.authorizer.refresh_token()
        self.transport.assert_no_requests()



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""
Geoscience Object API
=============


The Geoscience Object API enables technological integrations at the object level. It enables users to access and utilise their data across all products through a common and accessible data structure.

A Geoscience Object is a data structure that represents a concrete geological, geotechnical, or geophysical concept. Geoscience Objects can be referenced by their UUID or by a user-defined object path.

For more information on using the Geoscience Object API, see the [Geoscience Object API overview](/docs/guides/objects), or the API references here.


This code is generated from the OpenAPI specification for Geoscience Object API.
API version: 1.14.0
"""

from evo.common.connector import APIConnector
from evo.common.data import EmptyResponse, RequestMethod  # noqa: F401

from ..models import *  # noqa: F403

__all__ = ["DataApi"]


class DataApi:
    """API client for the Data endpoint.

    NOTE: This class is auto generated by OpenAPI Generator
    Ref: https://openapi-generator.tech

    Do not edit the class manually.

    :param connector: Client for communicating with the API.
    """

    def __init__(self, connector: APIConnector):
        self.connector = connector

    async def put_data_in_workspace(
        self,
        org_id: str,
        workspace_id: str,
        data_upload_request_body: list[DataUploadRequestBody],  # noqa: F405
        additional_headers: dict[str, str] | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> list[DataUploadResponseBody]:  # noqa: F405
        """Upload geoscience data

        Request to upload geoscience data.  The client provides names for the data they intend to upload and receives upload URLs for those that have not already been uploaded. This data `name` must a valid SHA256 hash or UUID. Calculating the SHA256 hash is recommended as it enables clients to avoid re-uploading duplicate data.  Each upload URL is secured with a SAS token and is valid for a limited time. If a URL has expired before the data upload is complete, simply repeat this request to obtain a new URL. Clients can use the new URL to continue uploading where they left off.  Multiple data names can be provided, with an upload URL generated for each data digest individually (up to a maximum of 32).

        :param org_id: The customer's organisation ID
            Format: `uuid`
            Example: `'b208a6c9-6881-4b97-b02d-acb5d81299bb'`
        :param workspace_id:
            Format: `uuid`
            Example: `'032806a8-dcd7-11ed-8d5c-00155d8f28b5'`
        :param data_upload_request_body:
            Example: `[endpoints.DataUploadRequestBody()]`
        :param additional_headers: (optional) Additional headers to send with the request.
        :param request_timeout: (optional) Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: Returns the result object.

        :raise evo.common.exceptions.BadRequestException: If the server responds with HTTP status 400.
        :raise evo.common.exceptions.UnauthorizedException: If the server responds with HTTP status 401.
        :raise evo.common.exceptions.ForbiddenException: If the server responds with HTTP status 403.
        :raise evo.common.exceptions.NotFoundException: If the server responds with HTTP status 404.
        :raise evo.common.exceptions.BaseTypedError: If the server responds with any other HTTP status between
            400 and 599, and the body of the response contains a descriptive `type` parameter.
        :raise evo.common.exceptions.EvoAPIException: If the server responds with any other HTTP status between 400
            and 599, and the body of the response does not contain a `type` parameter.
        :raise evo.common.exceptions.UnknownResponseError: For other HTTP status codes with no corresponding response
            type in `response_types_map`.
        """
        # Prepare the path parameters.
        _path_params = {
            "org_id": org_id,
            "workspace_id": workspace_id,
        }

        # Prepare the header parameters.
        _header_params = {
            "Content-Type": "application/json",
            "Accept": "application/json",
        }
        if additional_headers is not None:
            _header_params.update(additional_headers)

        # Define the collection formats.
        _collection_formats = {
            "DataUploadRequestBody": "",
        }

        _response_types_map = {
            "200": list[DataUploadResponseBody],  # noqa: F405
        }

        return await self.connector.call_api(
            method=RequestMethod.PUT,
            resource_path="/geoscience-object/orgs/{org_id}/workspaces/{workspace_id}/data",
            path_params=_path_params,
            header_params=_header_params,
            body=data_upload_request_body,
            collection_formats=_collection_formats,
            response_types_map=_response_types_map,
            request_timeout=request_timeout,
        )



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""
Block Model API
=============


    The Block Model API provides the ability to manage and report on block models in your Evo workspaces. Enable your
    product with Evo connected workflows by integrating with the Seequent Block Model API.

    The Block Model API supports a range of sub-blocking options, and both full and partial updates to block models,
    including updates of specified columns and/or sub-volumes. Block models are versioned, and the service supports
    reporting on the material content of current or previous versions, and comparing the content between versions of
    a block model.

    For more information on the Block Model API, see [Overview](/docs/blockmodel/overview), or the API references here.


This code is generated from the OpenAPI specification for Block Model API.
API version: 1.16.1
"""

from evo.common.connector import APIConnector
from evo.common.data import EmptyResponse, RequestMethod  # noqa: F401

from ..models import *  # noqa: F403

__all__ = ["MetadataApi"]


class MetadataApi:
    """API client for the Metadata endpoint.

    NOTE: This class is auto generated by OpenAPI Generator
    Ref: https://openapi-generator.tech

    Do not edit the class manually.

    :param connector: Client for communicating with the API.
    """

    def __init__(self, connector: APIConnector):
        self.connector = connector

    async def list_block_models(
        self,
        workspace_id: str,
        org_id: str,
        deleted: bool | None = None,
        offset: int | None = None,
        limit: int | None = None,
        additional_headers: dict[str, str] | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> PaginatedResponseBlockModel:  # noqa: F405
        """List all block models in a workspace

        Gets a paginated list of all block models in a workspace identified by `workspace_id`. By retrieving this list, you can obtain the `bm_id` value for a given model, which can then be used in subsequent API requests to upload or download data, along with performing other operations.  The list is ordered from the most recently updated block model to the least recently updated block model. By default, the list is ordered from the most recently updated block model to the least recently updated block model. If a `sort` query parameter is specified, the list will be ordered according to the specified field.  This endpoint is paginated, therefore by default this lists at most the first 50 block models. To get other block models, use the `offset` and `limit` query parameters to select the desired part of the list. An `offset` beyond the total number of block models in the workspace will result in an empty `results` list. The `limit` must be an integer from 1 to 100. The response includes `total`, which is the total number of block models within the list.  All workspace roles can use this endpoint.

        :param workspace_id: ID of the workspace this call is scoped to. Represented as a v4 UUID.
            Format: `uuid`
            Example: `'860be2f5-fe06-4c1b-ac8b-7d34d2b6d2ef'`
        :param org_id: ID of the organization this call is scoped to. Represented as a v4 UUID.
            Format: `uuid`
            Example: `'bf1a040c-8c58-4bc2-bec2-c5ae7de8bd84'`
        :param deleted: (optional)  An optional boolean parameter specifying whether to list only deleted block models.  - If set to `true`, the response will only list block models that were deleted via the 'Delete a block model' endpoint. This list will not include block models that are within deleted workspaces.  - By default, or if set to `false`, the response will only list non-deleted block models within non-deleted workspaces.
            Example: `False`
        :param offset: (optional) Index of the first item to return.
            Example: `0`
        :param limit: (optional) Maximum number of items to return from the list. Must be greater than 0 and less than or equal to 100.
            Example: `50`
        :param additional_headers: (optional) Additional headers to send with the request.
        :param request_timeout: (optional) Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: Returns the result object.

        :raise evo.common.exceptions.BadRequestException: If the server responds with HTTP status 400.
        :raise evo.common.exceptions.UnauthorizedException: If the server responds with HTTP status 401.
        :raise evo.common.exceptions.ForbiddenException: If the server responds with HTTP status 403.
        :raise evo.common.exceptions.NotFoundException: If the server responds with HTTP status 404.
        :raise evo.common.exceptions.BaseTypedError: If the server responds with any other HTTP status between
            400 and 599, and the body of the response contains a descriptive `type` parameter.
        :raise evo.common.exceptions.EvoAPIException: If the server responds with any other HTTP status between 400
            and 599, and the body of the response does not contain a `type` parameter.
        :raise evo.common.exceptions.UnknownResponseError: For other HTTP status codes with no corresponding response
            type in `response_types_map`.
        """
        # Prepare the path parameters.
        _path_params = {
            "workspace_id": workspace_id,
            "org_id": org_id,
        }

        # Prepare the query parameters.
        _query_params = {}
        if deleted is not None:
            _query_params["deleted"] = deleted
        if offset is not None:
            _query_params["offset"] = offset
        if limit is not None:
            _query_params["limit"] = limit

        # Prepare the header parameters.
        _header_params = {
            "Accept": "application/json",
        }
        if additional_headers is not None:
            _header_params.update(additional_headers)

        # Define the collection formats.
        _collection_formats = {}

        _response_types_map = {
            "200": PaginatedResponseBlockModel,  # noqa: F405
        }

        return await self.connector.call_api(
            method=RequestMethod.GET,
            resource_path="/blockmodel/orgs/{org_id}/workspaces/{workspace_id}/block-models",
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            collection_formats=_collection_formats,
            response_types_map=_response_types_map,
            request_timeout=request_timeout,
        )

    async def list_block_models_across_workspaces(
        self,
        org_id: str,
        deleted: bool | None = None,
        offset: int | None = None,
        limit: int | None = None,
        additional_headers: dict[str, str] | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> PaginatedResponseBlockModel:  # noqa: F405
        """List all block models in an organization

        Gets a paginated list of all block models across all workspaces that are accessible in the organization identified by `org_id`. By retrieving this list, you can obtain the `bm_id` value for a given model, which can then be used in subsequent API requests to upload or download data, along with performing other operations.  By default, the list is ordered from the most recently updated block model to the least recently updated block model. If a `sort` query parameter is specified, the list will be ordered according to the specified field.  This endpoint is paginated, therefore by default this lists at most the first 50 block models. To get other block models, use the `offset` and `limit` query parameters to select the desired part of the list. An `offset` beyond the total number of block models across all workspaces will result in an empty `results` list. The `limit` must be an integer from 1 to 100. The response includes `total`, which is the total number of block models within the list.  All workspace roles can use this endpoint.

        :param org_id: ID of the organization this call is scoped to. Represented as a v4 UUID.
            Format: `uuid`
            Example: `'bf1a040c-8c58-4bc2-bec2-c5ae7de8bd84'`
        :param deleted: (optional)  An optional boolean parameter specifying whether to list only deleted block models.  - If set to `true`, the response will only list block models that were deleted via the 'Delete a block model' endpoint. This list will not include block models that are within deleted workspaces.  - By default, or if set to `false`, the response will only list non-deleted block models within non-deleted workspaces.
            Example: `False`
        :param offset: (optional) Index of the first item to return.
            Example: `0`
        :param limit: (optional) Maximum number of items to return from the list. Must be greater than 0 and less than or equal to 100.
            Example: `50`
        :param additional_headers: (optional) Additional headers to send with the request.
        :param request_timeout: (optional) Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: Returns the result object.

        :raise evo.common.exceptions.BadRequestException: If the server responds with HTTP status 400.
        :raise evo.common.exceptions.UnauthorizedException: If the server responds with HTTP status 401.
        :raise evo.common.exceptions.ForbiddenException: If the server responds with HTTP status 403.
        :raise evo.common.exceptions.NotFoundException: If the server responds with HTTP status 404.
        :raise evo.common.exceptions.BaseTypedError: If the server responds with any other HTTP status between
            400 and 599, and the body of the response contains a descriptive `type` parameter.
        :raise evo.common.exceptions.EvoAPIException: If the server responds with any other HTTP status between 400
            and 599, and the body of the response does not contain a `type` parameter.
        :raise evo.common.exceptions.UnknownResponseError: For other HTTP status codes with no corresponding response
            type in `response_types_map`.
        """
        # Prepare the path parameters.
        _path_params = {
            "org_id": org_id,
        }

        # Prepare the query parameters.
        _query_params = {}
        if deleted is not None:
            _query_params["deleted"] = deleted
        if offset is not None:
            _query_params["offset"] = offset
        if limit is not None:
            _query_params["limit"] = limit

        # Prepare the header parameters.
        _header_params = {
            "Accept": "application/json",
        }
        if additional_headers is not None:
            _header_params.update(additional_headers)

        # Define the collection formats.
        _collection_formats = {}

        _response_types_map = {
            "200": PaginatedResponseBlockModel,  # noqa: F405
        }

        return await self.connector.call_api(
            method=RequestMethod.GET,
            resource_path="/blockmodel/orgs/{org_id}/block-models",
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            collection_formats=_collection_formats,
            response_types_map=_response_types_map,
            request_timeout=request_timeout,
        )

    async def retrieve_block_model(
        self,
        bm_id: str,
        workspace_id: str,
        org_id: str,
        deleted: bool | None = None,
        additional_headers: dict[str, str] | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> BlockModel:  # noqa: F405
        """Get a block model's metadata

        Gets the details of the block model for the given `bm_id`.  All workspace roles can use this endpoint.

        :param bm_id: ID of the block model this call is scoped to. Represented as a v4 UUID.
            Format: `uuid`
            Example: `'e3c277c2-edc6-4a7a-8380-251dd19231f2'`
        :param workspace_id: ID of the workspace this call is scoped to. Represented as a v4 UUID.
            Format: `uuid`
            Example: `'860be2f5-fe06-4c1b-ac8b-7d34d2b6d2ef'`
        :param org_id: ID of the organization this call is scoped to. Represented as a v4 UUID.
            Format: `uuid`
            Example: `'bf1a040c-8c58-4bc2-bec2-c5ae7de8bd84'`
        :param deleted: (optional)  An optional boolean parameter. It must be set to `true` when retrieving the block model for the given `bm_id` that has been deleted via the 'Delete a block model' endpoint. By default, or if set to `false`, an attempt to access a deleted block model will result in an error.
            Example: `False`
        :param additional_headers: (optional) Additional headers to send with the request.
        :param request_timeout: (optional) Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: Returns the result object.

        :raise evo.common.exceptions.BadRequestException: If the server responds with HTTP status 400.
        :raise evo.common.exceptions.UnauthorizedException: If the server responds with HTTP status 401.
        :raise evo.common.exceptions.ForbiddenException: If the server responds with HTTP status 403.
        :raise evo.common.exceptions.NotFoundException: If the server responds with HTTP status 404.
        :raise evo.common.exceptions.BaseTypedError: If the server responds with any other HTTP status between
            400 and 599, and the body of the response contains a descriptive `type` parameter.
        :raise evo.common.exceptions.EvoAPIException: If the server responds with any other HTTP status between 400
            and 599, and the body of the response does not contain a `type` parameter.
        :raise evo.common.exceptions.UnknownResponseError: For other HTTP status codes with no corresponding response
            type in `response_types_map`.
        """
        # Prepare the path parameters.
        _path_params = {
            "bm_id": bm_id,
            "workspace_id": workspace_id,
            "org_id": org_id,
        }

        # Prepare the query parameters.
        _query_params = {}
        if deleted is not None:
            _query_params["deleted"] = deleted

        # Prepare the header parameters.
        _header_params = {
            "Accept": "application/json",
        }
        if additional_headers is not None:
            _header_params.update(additional_headers)

        # Define the collection formats.
        _collection_formats = {}

        _response_types_map = {
            "200": BlockModel,  # noqa: F405
        }

        return await self.connector.call_api(
            method=RequestMethod.GET,
            resource_path="/blockmodel/orgs/{org_id}/workspaces/{workspace_id}/block-models/{bm_id}",
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            collection_formats=_collection_formats,
            response_types_map=_response_types_map,
            request_timeout=request_timeout,
        )



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""
Block Model API
=============


    The Block Model API provides the ability to manage and report on block models in your Evo workspaces. Enable your
    product with Evo connected workflows by integrating with the Seequent Block Model API.

    The Block Model API supports a range of sub-blocking options, and both full and partial updates to block models,
    including updates of specified columns and/or sub-volumes. Block models are versioned, and the service supports
    reporting on the material content of current or previous versions, and comparing the content between versions of
    a block model.

    For more information on the Block Model API, see [Overview](/docs/blockmodel/overview), or the API references here.


This code is generated from the OpenAPI specification for Block Model API.
API version: 1.16.1
"""

from evo.common.connector import APIConnector
from evo.common.data import EmptyResponse, RequestMethod

from ..models import *  # noqa: F403

__all__ = ["VersionsApi"]


class VersionsApi:
    """API client for the Versions endpoint.

    NOTE: This class is auto generated by OpenAPI Generator
    Ref: https://openapi-generator.tech

    Do not edit the class manually.

    :param connector: Client for communicating with the API.
    """

    def __init__(self, connector: APIConnector):
        self.connector = connector

    async def get_deltas_for_block_model(
        self,
        version_id: str,
        workspace_id: str,
        org_id: str,
        bm_id: str,
        delta_request_data: DeltaRequestData,  # noqa: F405
        additional_headers: dict[str, str] | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> DeltaResponseData:  # noqa: F405
        """Check for deltas

        Checks if there have been any changes to the block model within a given bounding box, between two specified versions.  Changes are searched for within a range of versions. The first version searched for changes is the version after `version_id`. The last version that is searched is the version that is specified by `end_version_id` (inclusive), or if `end_version_id` isn't specified, the latest version of the block model.  This request only checks for updates or deletions for specified columns. The columns are specified as list of `column_id` UUIDs or a wildcard (`\"*\"`), for all columns that exist within the `version_id` version (the starting version). Using column titles is not supported on this endpoint.  For this endpoint, changes that are considered are: - Addition of any new columns, regardless of the columns specified in this request - Deletions of any of the columns specified - Updates to any of the columns specified, within the given bounding box. When block models are updated, a bounding box that encloses all updated blocks is computed. Only updates that have a bounding box that intersects with the specified bounding box are included.  Column renames aren't considered as part of this and they do not count as a change.  If no changes are found, then the response will have a status code of 304, with no content in the response.  If there are changes found, then the `verbose` flag inside the request body changes the response. If `verbose` is false, then the response body will be empty. If `verbose` is true, then the response body will contain information about the difference in versions. For clarity, if no changes are found, then the `verbose` flag does nothing, as in that case the system will always return a 304 with no content.  All workspace roles can use this endpoint.

        :param version_id: ID of the version of the block model this call is scoped to. Represented as a v4 UUID.
            Format: `uuid`
            Example: `'3e9ce8de-f6ba-4920-8c6e-0882e90f0ed7'`
        :param workspace_id: ID of the workspace this call is scoped to. Represented as a v4 UUID.
            Format: `uuid`
            Example: `'860be2f5-fe06-4c1b-ac8b-7d34d2b6d2ef'`
        :param org_id: ID of the organization this call is scoped to. Represented as a v4 UUID.
            Format: `uuid`
            Example: `'bf1a040c-8c58-4bc2-bec2-c5ae7de8bd84'`
        :param bm_id: ID of the block model this call is scoped to. Represented as a v4 UUID.
            Format: `uuid`
            Example: `'e3c277c2-edc6-4a7a-8380-251dd19231f2'`
        :param delta_request_data:
            Example: `endpoints.DeltaRequestData()`
        :param additional_headers: (optional) Additional headers to send with the request.
        :param request_timeout: (optional) Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: Returns the result object.

        :raise evo.common.exceptions.BadRequestException: If the server responds with HTTP status 400.
        :raise evo.common.exceptions.UnauthorizedException: If the server responds with HTTP status 401.
        :raise evo.common.exceptions.ForbiddenException: If the server responds with HTTP status 403.
        :raise evo.common.exceptions.NotFoundException: If the server responds with HTTP status 404.
        :raise evo.common.exceptions.BaseTypedError: If the server responds with any other HTTP status between
            400 and 599, and the body of the response contains a descriptive `type` parameter.
        :raise evo.common.exceptions.EvoAPIException: If the server responds with any other HTTP status between 400
            and 599, and the body of the response does not contain a `type` parameter.
        :raise evo.common.exceptions.UnknownResponseError: For other HTTP status codes with no corresponding response
            type in `response_types_map`.
        """
        # Prepare the path parameters.
        _path_params = {
            "version_id": version_id,
            "workspace_id": workspace_id,
            "org_id": org_id,
            "bm_id": bm_id,
        }

        # Prepare the header parameters.
        _header_params = {
            "Content-Type": "application/json",
            "Accept": "application/json",
        }
        if additional_headers is not None:
            _header_params.update(additional_headers)

        # Define the collection formats.
        _collection_formats = {}

        _response_types_map = {
            "200": DeltaResponseData,  # noqa: F405
            "304": EmptyResponse,
        }

        return await self.connector.call_api(
            method=RequestMethod.POST,
            resource_path="/blockmodel/orgs/{org_id}/workspaces/{workspace_id}/block-models/{bm_id}/versions/{version_id}/delta",
            path_params=_path_params,
            header_params=_header_params,
            body=delta_request_data,
            collection_formats=_collection_formats,
            response_types_map=_response_types_map,
            request_timeout=request_timeout,
        )

    async def list_block_model_versions(
        self,
        workspace_id: str,
        org_id: str,
        bm_id: str,
        filter: str | None = None,
        offset: int | None = None,
        limit: int | None = None,
        additional_headers: dict[str, str] | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> PaginatedResponseWithUnitsVersion:  # noqa: F405
        """List version metadata for all versions of a block model

        Gets all versions of the block model `bm_id`.  The list is ordered from the newest version to the oldest version.  This endpoint is paginated, therefore by default this lists, at most, the first 50 versions. To get other versions, use the `offset` and `limit` query parameters to select the desired part of the list. An `offset` beyond the total number of versions for the block model will result in an empty `results` list. The `limit` must be an integer from 1 to 100. The response includes `total`, which is the total number of versions within the list. All workspace roles can use this endpoint.

        :param workspace_id: ID of the workspace this call is scoped to. Represented as a v4 UUID.
            Format: `uuid`
            Example: `'860be2f5-fe06-4c1b-ac8b-7d34d2b6d2ef'`
        :param org_id: ID of the organization this call is scoped to. Represented as a v4 UUID.
            Format: `uuid`
            Example: `'bf1a040c-8c58-4bc2-bec2-c5ae7de8bd84'`
        :param bm_id: ID of the block model this call is scoped to. Represented as a v4 UUID.
            Format: `uuid`
            Example: `'e3c277c2-edc6-4a7a-8380-251dd19231f2'`
        :param filter: (optional) Optional filter to apply to the results. `latest` will return only the latest version.
            Example: `'filter_example'`
        :param offset: (optional) Index of the first item to return.
            Example: `0`
        :param limit: (optional) Maximum number of items to return from the list. Must be greater than 0 and less than or equal to 100.
            Example: `50`
        :param additional_headers: (optional) Additional headers to send with the request.
        :param request_timeout: (optional) Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: Returns the result object.

        :raise evo.common.exceptions.BadRequestException: If the server responds with HTTP status 400.
        :raise evo.common.exceptions.UnauthorizedException: If the server responds with HTTP status 401.
        :raise evo.common.exceptions.ForbiddenException: If the server responds with HTTP status 403.
        :raise evo.common.exceptions.NotFoundException: If the server responds with HTTP status 404.
        :raise evo.common.exceptions.BaseTypedError: If the server responds with any other HTTP status between
            400 and 599, and the body of the response contains a descriptive `type` parameter.
        :raise evo.common.exceptions.EvoAPIException: If the server responds with any other HTTP status between 400
            and 599, and the body of the response does not contain a `type` parameter.
        :raise evo.common.exceptions.UnknownResponseError: For other HTTP status codes with no corresponding response
            type in `response_types_map`.
        """
        # Prepare the path parameters.
        _path_params = {
            "workspace_id": workspace_id,
            "org_id": org_id,
            "bm_id": bm_id,
        }

        # Prepare the query parameters.
        _query_params = {}
        if filter is not None:
            _query_params["filter"] = filter
        if offset is not None:
            _query_params["offset"] = offset
        if limit is not None:
            _query_params["limit"] = limit

        # Prepare the header parameters.
        _header_params = {
            "Accept": "application/json",
        }
        if additional_headers is not None:
            _header_params.update(additional_headers)

        # Define the collection formats.
        _collection_formats = {}

        _response_types_map = {
            "200": PaginatedResponseWithUnitsVersion,  # noqa: F405
        }

        return await self.connector.call_api(
            method=RequestMethod.GET,
            resource_path="/blockmodel/orgs/{org_id}/workspaces/{workspace_id}/block-models/{bm_id}/versions",
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            collection_formats=_collection_formats,
            response_types_map=_response_types_map,
            request_timeout=request_timeout,
        )

    async def query_block_model_latest_as_post(
        self,
        bm_id: str,
        workspace_id: str,
        org_id: str,
        query_criteria: QueryCriteria,  # noqa: F405
        additional_headers: dict[str, str] | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> QueryResult:  # noqa: F405
        """Start a block model data query

        Starts a query job against the block model identified by `bm_id`. This will produce a file that contains block model data for selected columns, and within the bounding box if provided.  This request will respond with a reflection of the query criteria, as well as a pollable `job_url`. The `job_url` should be polled with a GET request to wait for the `job_status` field of the job to become either `FAILED` or `COMPLETE`. When the `job_status` becomes `COMPLETE`, the response will look like this:  ``` {     \"job_status\": \"COMPLETE\",     \"payload\": {         \"download_url\": \"<download url>\"     } } ```  The `download_url` points to the output file of the query job, and can be downloaded using either a GET request or using the Azure SDK. The download URL that is generated every time the completed `job_url` is polled has a 30 minute time-to-live (TTL), so the download process must be started within 30 minutes. The download  itself can take longer than 30 minutes, as long as the connection remains open.  The downloaded file will either be an Apache Parquet file or a CSV file, depending on the value specified in the `file_format` field under `output_options`. The columns included within the file depend on the type of sub-blocking and the values specified in the `geometry_columns` and `columns` fields. The `geometry_columns` field determines whether the blocks within the output file are primarily identified by their coordinates or by their block indices.  If `geometry_columns` is set to \"indices\" (the default), then the first columns in the output file will be `i`, `j`, `k` followed by the sub-block index columns (if applicable). If `geometry_columns` is set to \"coordinates\", then the output file will contain the columns `x`, `y`, `z` for regular block models and `x`, `y`, `z`, `dx`, `dy`, `dz` for sub-blocked block models. These columns are referred to as the \"geometry\" columns.  The remaining columns within the output file are determined by the `columns` field. The `columns` field supports selecting columns by either their title or ID, and can also include a wildcard (`\"*\"`) placeholder, which will expand to all user columns, ordered alphabetically by title, in a case-insensitive manner. Please note that the wildcard does not cover the system column `version_id`. To include `version_id` in the output file, it must also be explicitly specified in the `columns` field alongside the wildcard. The order of columns in the output file will match the order in the `columns` field. Columns that are part of the initial \"geometry\" columns will be ignored if specified.  Examples: Given a model has the following user columns: `A`(with a column ID of `d718abe4-56a5-4e27-ad51-813e69eb8aac`), `B`, and `C`. The table below shows the output file columns for different model types and parameters:  | Model Type | `geometry_columns` field | `columns` field | Output File Columns | |------------|--------------------------|-----------------|---------------------| | Regular           | \"indices\"         | `[\"d718abe4-56a5-4e27-ad51-813e69eb8aac\", \"B\"]`        | `i`, `j`, `k`, `A`, `B` | | Regular           | \"coordinates\"     | `[]`                | `x`, `y`, `z` | | Fully sub-blocked | \"indices\"         | `[\"*\"]`             | `i`, `j`, `k`, `sidx`, `A`, `B`, `C` | | Flexible          | \"coordinates\"     | `[\"*\", \"start_si\"]` | `x`, `y`, `z`, `dx`, `dy`, `dz`, `A`, `B`, `C`, `start_si` |   The column headers (header row values in CSV files and schema field names in Parquet files) will either be the name of the columns or the column ID, depending on the value specified in the `column_headers` field.  Example: for a fully sub-blocked model with a `columns` field: `[\"col_1\", \"dx\", \"x\", \"z\", \"d635868c-24ae-4ebe-9358-acb433b3c8cb\"]` (where `d635868c-24ae-4ebe-9358-acb433b3c8cb` is the column ID of `col_2`) and `column_headers`: `name`, the output file will contain the following columns in order: [`i`, `j`, `k`, `sidx`, `col_1`, `dx`, `x`, `z`, `col_2`].  For CSV files, the delimiter that separate values in the output file can also be configured using the `delimiter` field.  For Parquet files, the output file will have the following characteristics: - Row group size of 100,000 - Compressed using Zstd - Parquet data page version set to 1.0 and Parquet version set to 2.6  This request may be cached, so the `job_url` may point to an already completed job.  All workspace roles can use this endpoint.

        :param bm_id: ID of the block model this call is scoped to. Represented as a v4 UUID.
            Format: `uuid`
            Example: `'e3c277c2-edc6-4a7a-8380-251dd19231f2'`
        :param workspace_id: ID of the workspace this call is scoped to. Represented as a v4 UUID.
            Format: `uuid`
            Example: `'860be2f5-fe06-4c1b-ac8b-7d34d2b6d2ef'`
        :param org_id: ID of the organization this call is scoped to. Represented as a v4 UUID.
            Format: `uuid`
            Example: `'bf1a040c-8c58-4bc2-bec2-c5ae7de8bd84'`
        :param query_criteria:
            Example: `endpoints.QueryCriteria()`
        :param additional_headers: (optional) Additional headers to send with the request.
        :param request_timeout: (optional) Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: Returns the result object.

        :raise evo.common.exceptions.BadRequestException: If the server responds with HTTP status 400.
        :raise evo.common.exceptions.UnauthorizedException: If the server responds with HTTP status 401.
        :raise evo.common.exceptions.ForbiddenException: If the server responds with HTTP status 403.
        :raise evo.common.exceptions.NotFoundException: If the server responds with HTTP status 404.
        :raise evo.common.exceptions.BaseTypedError: If the server responds with any other HTTP status between
            400 and 599, and the body of the response contains a descriptive `type` parameter.
        :raise evo.common.exceptions.EvoAPIException: If the server responds with any other HTTP status between 400
            and 599, and the body of the response does not contain a `type` parameter.
        :raise evo.common.exceptions.UnknownResponseError: For other HTTP status codes with no corresponding response
            type in `response_types_map`.
        """
        # Prepare the path parameters.
        _path_params = {
            "bm_id": bm_id,
            "workspace_id": workspace_id,
            "org_id": org_id,
        }

        # Prepare the header parameters.
        _header_params = {
            "Content-Type": "application/json",
            "Accept": "application/json",
        }
        if additional_headers is not None:
            _header_params.update(additional_headers)

        # Define the collection formats.
        _collection_formats = {}

        _response_types_map = {
            "200": QueryResult,  # noqa: F405
        }

        return await self.connector.call_api(
            method=RequestMethod.POST,
            resource_path="/blockmodel/orgs/{org_id}/workspaces/{workspace_id}/block-models/{bm_id}/blocks",
            path_params=_path_params,
            header_params=_header_params,
            body=query_criteria,
            collection_formats=_collection_formats,
            response_types_map=_response_types_map,
            request_timeout=request_timeout,
        )

    async def retrieve_block_model_version(
        self,
        version_id: str,
        workspace_id: str,
        org_id: str,
        bm_id: str,
        include_changes: bool | None = None,
        additional_headers: dict[str, str] | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> VersionWithChanges | Version:  # noqa: F405
        """Request version metadata for a specific block model version

        Gets the version of the block model `bm_id` identified by the `version_id` UUID.  If the `include_changes` query parameter is set to `true`, then the response will include the changes made in the update. Returned object will be of type `VersionWithChanges`, this object extends the `Version` object to include the field `changes`. The changes field contains the `VersionChanges` object.  The fields of the `VersionChanges` object are as follows: - blocks_uploaded: The number of blocks (rows) present in the update file if one was used for the update, null otherwise. - total_blocks: The total number of blocks in the block model. Will remain static for regular models, will change for sub-blocked models when the sub-block count changes due to geometry change updates. - update_type: Whether the update that created the Version was of type `merge` or `replace`; field contains null if the update did not update any blocks. - columns: - columns->new: List of columns (specified by title) created in the update that created the Version. Empty if the update did not create any columns. - columns->deleted: List of columns (specified by title) deleted in the update that created the Version. Empty if the update did not delete any columns. - columns->updated: List of columns (specified by title) updated in the update that created the Version. Empty if the update did not update any columns. - columns->renamed: List of the `ColumnRename` object, denoting which columns were renamed in the update the created the version. Empty if the update did not rename any columns. `ColumnRename` contains the fields: (`old_title`, `new_title`). - columns->metadata_updated: List of the `UpdateMetadataLite` object, containing column metadata changes if they were made in the update. This includes the `title` of the column, the new title, and updated unit ID, respective of what was modified in the update. Empty if the update did not update any column metadata.   All workspace roles can use this endpoint.

        :param version_id: ID of the version of the block model this call is scoped to. Represented as a v4 UUID.
            Format: `uuid`
            Example: `'3e9ce8de-f6ba-4920-8c6e-0882e90f0ed7'`
        :param workspace_id: ID of the workspace this call is scoped to. Represented as a v4 UUID.
            Format: `uuid`
            Example: `'860be2f5-fe06-4c1b-ac8b-7d34d2b6d2ef'`
        :param org_id: ID of the organization this call is scoped to. Represented as a v4 UUID.
            Format: `uuid`
            Example: `'bf1a040c-8c58-4bc2-bec2-c5ae7de8bd84'`
        :param bm_id: ID of the block model this call is scoped to. Represented as a v4 UUID.
            Format: `uuid`
            Example: `'e3c277c2-edc6-4a7a-8380-251dd19231f2'`
        :param include_changes: (optional) Whether or not to include the changes made in the update that created the specified block model version.
            Example: `False`
        :param additional_headers: (optional) Additional headers to send with the request.
        :param request_timeout: (optional) Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: Returns the result object.

        :raise evo.common.exceptions.BadRequestException: If the server responds with HTTP status 400.
        :raise evo.common.exceptions.UnauthorizedException: If the server responds with HTTP status 401.
        :raise evo.common.exceptions.ForbiddenException: If the server responds with HTTP status 403.
        :raise evo.common.exceptions.NotFoundException: If the server responds with HTTP status 404.
        :raise evo.common.exceptions.BaseTypedError: If the server responds with any other HTTP status between
            400 and 599, and the body of the response contains a descriptive `type` parameter.
        :raise evo.common.exceptions.EvoAPIException: If the server responds with any other HTTP status between 400
            and 599, and the body of the response does not contain a `type` parameter.
        :raise evo.common.exceptions.UnknownResponseError: For other HTTP status codes with no corresponding response
            type in `response_types_map`.
        """
        # Prepare the path parameters.
        _path_params = {
            "version_id": version_id,
            "workspace_id": workspace_id,
            "org_id": org_id,
            "bm_id": bm_id,
        }

        # Prepare the query parameters.
        _query_params = {}
        if include_changes is not None:
            _query_params["include_changes"] = include_changes

        # Prepare the header parameters.
        _header_params = {
            "Accept": "application/json",
        }
        if additional_headers is not None:
            _header_params.update(additional_headers)

        # Define the collection formats.
        _collection_formats = {}

        _response_types_map = {
            "200": VersionWithChanges | Version,  # noqa: F405
        }

        return await self.connector.call_api(
            method=RequestMethod.GET,
            resource_path="/blockmodel/orgs/{org_id}/workspaces/{workspace_id}/block-models/{bm_id}/versions/{version_id}",
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            collection_formats=_collection_formats,
            response_types_map=_response_types_map,
            request_timeout=request_timeout,
        )



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import annotations

import hashlib
import os
from pathlib import Path
from tempfile import NamedTemporaryFile
from types import EllipsisType

import pyarrow as pa
import pyarrow.parquet as pq
from typing_extensions import deprecated

import evo.logging
from evo.common.exceptions import StorageFileNotFoundError

from ..exceptions import SchemaValidationError, TableFormatError

logger = evo.logging.getLogger("object.tables")

__all__ = [
    "ArrowTableFormat",
    "BaseTableFormat",
    "KnownTableFormat",
]

# Block size to limit memory use while calculating the sha256 digest of a file.
_DIGEST_BLOCK_SIZE = 4 * 1024 * 1024


class _ColumnFormat:
    def __init__(self, format_spec: pa.DataType | str):
        if isinstance(format_spec, str):
            self._type = self._get_data_type(format_spec)
            self._format_id = format_spec
        else:
            self._type = format_spec
            self._format_id = self._get_format_id(format_spec)

    @staticmethod
    def _get_data_type(format_id: str) -> pa.DataType:
        match format_id:
            case "float64":
                return pa.float64()
            case "uint8":
                return pa.uint8()
            case "uint32":
                return pa.uint32()
            case "uint64":
                return pa.uint64()
            case "int32":
                return pa.int32()
            case "int64":
                return pa.int64()
            case "bool":
                return pa.bool_()
            case "string":
                return pa.string()
            case "timestamp":
                return pa.timestamp("us", tz="UTC")
            case unknown_format:
                raise TypeError(f"Unsupported column type '{unknown_format}'")

    @staticmethod
    def _get_format_id(data_type: pa.DataType) -> str:
        match str(data_type):
            case "double":
                return "float64"
            case "uint8":
                return "uint8"
            case "uint32":
                return "uint32"
            case "uint64":
                return "uint64"
            case "int32":
                return "int32"
            case "int64":
                return "int64"
            case "bool":
                return "bool"
            case "string":
                return "string"
            case "timestamp[us, tz=UTC]":
                return "timestamp"
            case unknown_format:
                raise TypeError(f"Unsupported column type '{unknown_format}' from {data_type!r}")

    @property
    def id(self) -> str:
        return self._format_id

    @property
    def type(self) -> pa.DataType:
        return self._type


class BaseTableFormat:
    """Base type for comparing table formats"""

    def __init__(self, name: str, columns: list[pa.DataType | str | EllipsisType]) -> None:
        """
        :param name: The display name for this format.
        :param columns: A list of column data types in this format. A single column data type followed by Ellipsis
            ("...") indicates a multidimensional format.
        """
        self._name = name
        self._columns: list[_ColumnFormat] = []
        self._multi_dimensional = False
        for column_type in columns:
            if column_type is Ellipsis:
                self._multi_dimensional = True
            elif self._multi_dimensional:
                raise AssertionError("Found new type specification after Ellipsis")
            else:
                self._columns.append(_ColumnFormat(format_spec=column_type))

        if self._multi_dimensional and len(self._columns) != 1:
            raise AssertionError("Multidimensional formats must only have one column type")
        elif len(self._columns) < 1:
            raise AssertionError("Table formats must have at least one column")

        # Calculate data type string in constructor to mitigate cost of `_check_format()`.
        column_types = [column_format.id for column_format in self._columns]
        unique_types = set(column_types)
        if len(unique_types) == 1:
            self._data_type = unique_types.pop()
        else:
            self._data_type = "/".join(column_types)

    @property
    def name(self) -> str:
        """Format display name"""
        return self._name

    @property
    def width(self) -> int:
        """The number of columns in this table format"""
        return len(self._columns)

    @property
    def data_type(self) -> str:
        """The column data type(s) in this table format"""
        return self._data_type


class ArrowTableFormat(BaseTableFormat):
    """Specialised table format type that can be generated from a pyarrow table"""

    @classmethod
    def from_schema(cls, pa_schema: pa.Schema) -> ArrowTableFormat:
        """Generate an ArrowTableFormat instance that represents the structure of the provided table schema.

        :param pa_schema: Table schema to generate a format representation for.

        :return: A table format that can be used to compare against known formats.
        """
        return ArrowTableFormat(name=f"{cls.__name__}<{'/'.join(pa_schema.names)}>", columns=pa_schema.types)


class KnownTableFormat(BaseTableFormat):
    """A definition of a known table format that matches a Geoscience Object Schema model type"""

    def __init__(self, name: str, columns: list[pa.DataType | EllipsisType], field_names: list[str] | None) -> None:
        """
        :param name: The display name for this format.
        :param columns: A list of column data types in this format. A single column data type followed by Ellipsis
            ("...") indicates a multidimensional format.
        :param field_names: The field names that are included when saving a table in this format.
        """
        super().__init__(name, columns)
        if field_names is None:
            self._field_names = None
        else:
            self._field_names = frozenset(field_names)

    def _check_format(self, other: BaseTableFormat) -> None:
        """Check whether another table format satisfies this format.

        :param other: The other table format to compare against.

        :raises TableFormatError: If the other format does not satisfy this format.
        """
        if not self._multi_dimensional and self.width != other.width:
            raise TableFormatError(
                f"Column count ({other.width}) does not match expectation ({self.width}) for {self.name}"
            )

        expected_data_type = self.data_type
        actual_data_type = other.data_type
        if expected_data_type != actual_data_type:
            raise TableFormatError(
                f"{self.name} requires data_type='{expected_data_type}' but the provided table has '{actual_data_type}'"
            )

    def is_provided_by(self, other: BaseTableFormat) -> bool:
        """Test whether the other format meets the requirements of this format

        :param other: The other format to test. Usually generated from actual data.

        :return: True if the other format meets the requirements of this format.
        """
        try:
            self._check_format(other)
        except TableFormatError:
            return False
        else:
            return True

    @staticmethod
    def _get_file_digest(file_path: Path) -> str:
        """Get the sha256 digest of a file

        :param file_path: The path of the target file.

        :return: The sha256 digest of the target file.
        """
        with file_path.open("rb") as file:
            sha256_digest = hashlib.sha256()
            while block := file.read(_DIGEST_BLOCK_SIZE):
                sha256_digest.update(block)
        return sha256_digest.hexdigest()

    @classmethod
    def _save_table_as_parquet(cls, table: pa.Table, destination: Path) -> str:
        """Save a table in parquet format.

        :param table: The table to save to parquet file.
        :param destination: A local directory to save the parquet file in.

        :return: The sha256 digest of the saved parquet file, which is also the filename.

        :raises StorageFileNotFoundError: If the destination does not exist or is not a directory.
        """
        if not destination.is_dir():
            raise StorageFileNotFoundError(f"'{destination}' is not a directory")

        try:  # Write temporary file.
            with NamedTemporaryFile(delete=False, dir=destination) as tmp_file:
                logger.debug(f"Writing temporary file {tmp_file.name}")
                pq.write_table(table=table, where=tmp_file, version="2.4", compression="gzip")
            tmp_path = Path(tmp_file.name).resolve()
            logger.debug(f"Calculating file digest for {tmp_file.name}")
            data_ref = cls._get_file_digest(tmp_path)
        except BaseException:
            logger.error(f"Removing temporary file '{tmp_file.name}' due to an unhandled exception", exc_info=True)
            os.unlink(tmp_file.name)
            raise

        new_path = destination / data_ref
        try:  # Rename temporary file using sha256 digest.
            logger.debug(f"Renaming {tmp_path.name} to {new_path.name}")
            tmp_path.replace(new_path)
        except PermissionError:  # When the same file is already opened somewhere else.
            os.unlink(tmp_file.name)
            if new_path.exists():
                logger.info(f"File '{new_path.name}' already exists")
            else:
                logger.error(f"Failed to rename file '{tmp_file.name}' due to an unhandled exception", exc_info=True)
                raise

        return data_ref

    def save_table(self, table: pa.Table, destination: Path) -> dict:
        """Save a pyarrow table in parquet format and return a GO model of the table metadata.

        :param table: The table to save in parquet format.
        :param destination: The directory to save the parquet file in.

        :return: A dictionary representing the Geoscience Object Schema that corresponds to this format.

        :raises TableFormatError: If the provided table does not match this format.
        :raises StorageFileNotFoundError: If the destination does not exist or is not a directory.
        """
        logger.debug(f"Saving table in folder {destination}")
        other_format = ArrowTableFormat.from_schema(table.schema)
        try:
            self._check_format(other_format)
            data_id = self._save_table_as_parquet(table, destination)
        except (TableFormatError, StorageFileNotFoundError) as error:
            logger.error(f"Could not save table because {error}")
            raise

        try:
            # The following fields are all those expected to be needed by table info types.
            data_dict = {
                "data": data_id,
                "length": table.num_rows,
                "width": table.num_columns,
                "data_type": self.data_type,
            }
            if len(self._columns) == 2:  # For lookup table special case.
                data_dict["keys_data_type"] = self._columns[0].id
                data_dict["values_data_type"] = self._columns[1].id

            if self._field_names is not None:
                data_dict = {field: value for field, value in data_dict.items() if field in self._field_names}

            return data_dict

        except Exception as error:
            logger.error(error, exc_info=True)
            parquet_file = destination / data_id
            try:
                parquet_file.unlink(missing_ok=True)
            except Exception:
                pass
            raise

    @classmethod
    def from_table_info(cls, table_info: dict) -> KnownTableFormat:
        """Generate a KnownTableFormat instance that represents the provided table info.

        :param table_info: Table info to generate a format representation for.

        :return: A table format that can be used to compare against other formats.
        """
        # "data" isn't that useful as a name, but it should always exist.
        type_name = table_info.get("schema") or table_info.get("data") or "Unknown"
        n_columns = table_info.get("width", 1)

        if "data_type" in table_info:
            data_type = table_info["data_type"]
        elif "keys_data_type" in table_info and "values_data_type" in table_info:
            data_type = "{keys_data_type}/{values_data_type}".format(**table_info)
            n_columns = 2
        else:
            raise TableFormatError(f"Unknown data type for '{type_name}' table")

        columns = data_type.split("/")
        if n_columns > 1 and len(columns) == 1:
            columns *= n_columns

        if n_columns != len(columns):
            raise TableFormatError(f"Unable to determine column types for '{type_name}' with data type '{data_type}'")

        return KnownTableFormat(name=type_name, columns=columns, field_names=table_info.get("field_names"))

    @classmethod
    @deprecated("KnownTableFormat.load_table is deprecated, use evo.objects.parquet.ParquetLoader instead")
    def load_table(cls, table_info: dict, source: Path) -> pa.Table:
        """Load parquet data as a pyarrow.Table and verify the format against the provided table info.

        The parquet metadata will be used to make sure the file contents matches the expected format before the table
        is read into memory.

        :param table_info: The table info that defines the expected format. The model's `data` field must be the
            filename of the parquet file, which must exist in `source`.
        :param source: The source directory to read the parquet file from.

        :return: A pyarrow table loaded directly from the parquet file.

        :raises StorageFileNotFoundError: If the parquet file does not exist or is not a file.
        :raises TableFormatError: If the parquet data does not match the expected format.
        :raises SchemaValidationError: If the parquet file has a different number of rows than expected.
        """
        parquet_file = source / str(table_info["data"])
        logger.debug(f"Reading parquet data from {parquet_file}")
        try:
            with pa.OSFile(str(parquet_file), mode="r") as parquet_file_reader:
                parquet_data = pq.ParquetFile(parquet_file_reader)

                logger.debug("Checking parquet data format")
                expected_format = KnownTableFormat.from_table_info(table_info)
                actual_format = ArrowTableFormat.from_schema(parquet_data.schema_arrow)
                cls._check_format(expected_format, actual_format)

                logger.debug("Checking parquet data length")
                expected_length = table_info["length"]
                actual_length = parquet_data.metadata.num_rows
                if expected_length != actual_length:
                    raise SchemaValidationError(
                        f"Row count ({actual_length}) does not match expectation ({expected_length})"
                    )

                logger.debug("Parquet metadata checks succeeded")
                return parquet_data.read()

        except FileNotFoundError:
            raise StorageFileNotFoundError(f"Could not find data for table with id '{parquet_file.name}'") from None

        except (TableFormatError, SchemaValidationError) as error:
            logger.error(f"Could not load table because {error}")
            raise



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import sys
from typing import Generic, TypeAlias, TypeVar

if sys.version_info >= (3, 12):
    from typing import NotRequired, TypedDict
else:
    from typing_extensions import NotRequired, TypedDict

__all__ = ["ArrayTableInfo", "AttributeInfo", "CategoryInfo", "LookupTableInfo", "TableInfo"]


class _BaseTableInfo(TypedDict):
    data: str
    length: int


class ArrayTableInfo(_BaseTableInfo):
    """Metadata for a non-lookup table.

    The 'data' field contains the reference to the blob where the table data is stored.

    The 'length', 'width', and 'data_type' fields describe the structure of the table.
    """

    data_type: str
    width: NotRequired[int]


class LookupTableInfo(_BaseTableInfo):
    """Metadata for lookup table, which is used to define categories.

    The 'data' field contains the reference to the blob where the table data is stored.

    The 'length', 'width', and 'data_type' fields describe the structure of the table.
    """

    keys_data_type: str
    values_data_type: str


TableInfo: TypeAlias = ArrayTableInfo | LookupTableInfo


class CategoryInfo(TypedDict):
    """Metadata for category tables.

    In Geoscience Object Schemas, categories are defined by an indices array(values) and a lookup table (table).
    """

    table: LookupTableInfo
    values: TableInfo


T = TypeVar("T")


class _Nan(TypedDict, Generic[T]):
    values: list[T]


class NanCategorical(_Nan[int]):
    """Metadata for representing 'not a number' (NaN) values in categorical/integer attributes.

    In addition to supporting null values within certain tables, additional 'not a number' (NaN) values can be defined,
    which should be interpreted as 'not a number' (NaN).
    """


class NanContinuous(_Nan[float]):
    """Metadata for representing 'not a number' (NaN) values in continuous attributes.

    In addition to supporting null values within certain tables, additional 'not a number' (NaN) values can be defined,
    which should be interpreted as 'not a number' (NaN).
    """


class AttributeInfo(TypedDict):
    """Metadata for attributes."""

    name: str
    key: NotRequired[str]
    nan_description: NotRequired[NanCategorical | NanContinuous]
    values: ArrayTableInfo
    table: NotRequired[LookupTableInfo]



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from evo.common.test_tools import MockResponse
from evo.oauth import ClientCredentialsAuthorizer
from evo.oauth.data import AccessToken, EvoScopes

from ._helpers import CLIENT_ID, CLIENT_SECRET, TestWithOAuthConnector, get_access_token


class TestClientCredentialsAuthorizer(TestWithOAuthConnector):
    def setUp(self):
        super().setUp()
        self.authorizer = ClientCredentialsAuthorizer(oauth_connector=self.connector, scopes=EvoScopes.all_evo)
        self.first_token = get_access_token(access_token="first_token")
        self.second_token = get_access_token(access_token="second_token")

        self.transport.request.side_effect = [
            MockResponse(status_code=200, content=self.first_token.model_dump_json(by_alias=True, exclude_unset=True)),
            MockResponse(
                status_code=200,
                content=self.second_token.model_dump_json(by_alias=True, exclude_unset=True),
            ),
        ]

    def assert_token_equals(self, expected_token: AccessToken | None) -> None:
        """Check that the access token is as expected."""
        actual_token: AccessToken | None = self.authorizer._BaseAuthorizer__token
        self.assertEqual(expected_token, actual_token)

    async def test_authorize(self):
        await self.authorizer.authorize()
        self.assert_token_equals(self.first_token)
        self.assert_fetched_token(
            grant_type="client_credentials",
            scope="evo.blocksync evo.discovery evo.file evo.object evo.workspace",
            client_id=CLIENT_ID,
            client_secret=CLIENT_SECRET,
        )

    async def test_refresh_token(self):
        await self.authorizer.authorize()

        result = await self.authorizer.refresh_token()

        self.assertTrue(result)
        self.assert_fetched_token(
            grant_type="client_credentials",
            scope="evo.blocksync evo.discovery evo.file evo.object evo.workspace",
            client_id=CLIENT_ID,
            client_secret=CLIENT_SECRET,
        )
        self.assert_token_equals(self.second_token)

    async def test_get_default_headers(self):
        await self.authorizer.authorize()

        headers = await self.authorizer.get_default_headers()

        self.assertEqual(headers, {"Authorization": "Bearer first_token"})

    async def test_get_default_headers_authorize_not_explicitly_called(self):
        headers = await self.authorizer.get_default_headers()

        self.assertEqual(headers, {"Authorization": "Bearer first_token"})



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import unittest

from evo.oauth.authorizer import AccessTokenAuthorizer


class TestAccessTokenAuthorizer(unittest.IsolatedAsyncioTestCase):
    def setUp(self) -> None:
        self.authorizer = AccessTokenAuthorizer(access_token="abc-123")

    async def test_get_default_headers(self) -> None:
        headers = await self.authorizer.get_default_headers()
        assert headers == {"Authorization": "Bearer abc-123"}

    async def test_refresh_token(self) -> None:
        assert not await self.authorizer.refresh_token()



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""
Block Model API
=============


    The Block Model API provides the ability to manage and report on block models in your Evo workspaces. Enable your
    product with Evo connected workflows by integrating with the Seequent Block Model API.

    The Block Model API supports a range of sub-blocking options, and both full and partial updates to block models,
    including updates of specified columns and/or sub-volumes. Block models are versioned, and the service supports
    reporting on the material content of current or previous versions, and comparing the content between versions of
    a block model.

    For more information on the Block Model API, see [Overview](/docs/blockmodel/overview), or the API references here.


This code is generated from the OpenAPI specification for Block Model API.
API version: 1.16.1
"""

from evo.common.connector import APIConnector
from evo.common.data import EmptyResponse, RequestMethod  # noqa: F401

from ..models import *  # noqa: F403

__all__ = ["JobsApi"]


class JobsApi:
    """API client for the Jobs endpoint.

    NOTE: This class is auto generated by OpenAPI Generator
    Ref: https://openapi-generator.tech

    Do not edit the class manually.

    :param connector: Client for communicating with the API.
    """

    def __init__(self, connector: APIConnector):
        self.connector = connector

    async def get_job_status(
        self,
        job_id: str,
        workspace_id: str,
        org_id: str,
        bm_id: str,
        additional_headers: dict[str, str] | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> JobResponse:  # noqa: F405
        """Request job status/result

        Retrieves a given job.  A job represents a background activity in the BlockSync compute workers, such as performing a query or updating a block model based on an uploaded file. This endpoint indicates the job's status, and once it has completed or failed, details about the result are provided in a `payload` property.  All workspace roles can use this endpoint.

        :param job_id: ID of the job this call is scoped to. Represented as a v4 UUID.
            Format: `uuid`
            Example: `'ff0b8078-c7ec-41ab-87fb-f3eb32812b32'`
        :param workspace_id: ID of the workspace this call is scoped to. Represented as a v4 UUID.
            Format: `uuid`
            Example: `'860be2f5-fe06-4c1b-ac8b-7d34d2b6d2ef'`
        :param org_id: ID of the organization this call is scoped to. Represented as a v4 UUID.
            Format: `uuid`
            Example: `'bf1a040c-8c58-4bc2-bec2-c5ae7de8bd84'`
        :param bm_id: ID of the block model this call is scoped to. Represented as a v4 UUID.
            Format: `uuid`
            Example: `'e3c277c2-edc6-4a7a-8380-251dd19231f2'`
        :param additional_headers: (optional) Additional headers to send with the request.
        :param request_timeout: (optional) Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: Returns the result object.

        :raise evo.common.exceptions.BadRequestException: If the server responds with HTTP status 400.
        :raise evo.common.exceptions.UnauthorizedException: If the server responds with HTTP status 401.
        :raise evo.common.exceptions.ForbiddenException: If the server responds with HTTP status 403.
        :raise evo.common.exceptions.NotFoundException: If the server responds with HTTP status 404.
        :raise evo.common.exceptions.BaseTypedError: If the server responds with any other HTTP status between
            400 and 599, and the body of the response contains a descriptive `type` parameter.
        :raise evo.common.exceptions.EvoAPIException: If the server responds with any other HTTP status between 400
            and 599, and the body of the response does not contain a `type` parameter.
        :raise evo.common.exceptions.UnknownResponseError: For other HTTP status codes with no corresponding response
            type in `response_types_map`.
        """
        # Prepare the path parameters.
        _path_params = {
            "job_id": job_id,
            "workspace_id": workspace_id,
            "org_id": org_id,
            "bm_id": bm_id,
        }

        # Prepare the header parameters.
        _header_params = {
            "Accept": "application/json",
        }
        if additional_headers is not None:
            _header_params.update(additional_headers)

        # Define the collection formats.
        _collection_formats = {}

        _response_types_map = {
            "200": JobResponse,  # noqa: F405
        }

        return await self.connector.call_api(
            method=RequestMethod.GET,
            resource_path="/blockmodel/orgs/{org_id}/workspaces/{workspace_id}/block-models/{bm_id}/jobs/{job_id}",
            path_params=_path_params,
            header_params=_header_params,
            collection_formats=_collection_formats,
            response_types_map=_response_types_map,
            request_timeout=request_timeout,
        )



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import json
from unittest import mock
from uuid import UUID

from evo.common import HealthCheckType, RequestMethod
from evo.common.test_tools import BASE_URL, MockResponse, TestHTTPHeaderDict, TestWithConnector, utc_datetime
from evo.workspaces import (
    BasicWorkspace,
    OrderByOperatorEnum,
    ServiceUser,
    User,
    UserRole,
    Workspace,
    WorkspaceAPIClient,
    WorkspaceOrderByEnum,
    WorkspaceRole,
)

from ..data import load_test_data

ORG_UUID = UUID(int=0)
USER_ID = UUID(int=2)
BASE_PATH = f"/workspace/orgs/{ORG_UUID}"

TEST_USER = ServiceUser(id=USER_ID, name="Test User", email="test.user@unit.test")


def _test_workspace(ws_id: UUID, name: str) -> Workspace:
    """Factory method to create test workspace objects."""
    return Workspace(
        id=ws_id,
        display_name=name.title(),
        description=name.lower(),
        user_role=WorkspaceRole.owner,
        org_id=ORG_UUID,
        hub_url=BASE_URL,
        created_at=utc_datetime(2020, 1, 1),
        created_by=TEST_USER,
        updated_at=utc_datetime(2020, 1, 1),
        updated_by=TEST_USER,
    )


def _test_basic_workspace(ws_id: UUID, name: str) -> BasicWorkspace:
    """Factory method to create test basic workspace objects."""
    return BasicWorkspace(
        id=ws_id,
        display_name=name.title(),
    )


TEST_WORKSPACE_A = _test_workspace(UUID(int=0xA), "Test Workspace A")
TEST_WORKSPACE_B = _test_workspace(UUID(int=0xB), "Test Workspace B")
TEST_WORKSPACE_C = _test_workspace(UUID(int=0xC), "Test Workspace C")
TEST_BASIC_WORKSPACE_A = _test_basic_workspace(UUID(int=0xA), "Test Workspace A")
TEST_BASIC_WORKSPACE_B = _test_basic_workspace(UUID(int=0xB), "Test Workspace B")
TEST_BASIC_WORKSPACE_C = _test_basic_workspace(UUID(int=0xC), "Test Workspace C")


class TestWorkspaceClient(TestWithConnector):
    def setUp(self) -> None:
        super().setUp()
        self.workspace_client = WorkspaceAPIClient(connector=self.connector, org_id=ORG_UUID)

    async def test_get_service_health(self) -> None:
        with mock.patch("evo.workspaces.client.get_service_health") as mock_get_service_health:
            await self.workspace_client.get_service_health()
        mock_get_service_health.assert_called_once_with(self.connector, "workspace", check_type=HealthCheckType.FULL)

    def _empty_content(self) -> str:
        data = """{"results": [], "links": {"first": "http://firstlink", "last": "http://lastlink",
                "next": null, "previous": null, "count": 0, "total": 0}}"""
        return data

    async def test_list_workspaces_default_args(self):
        with self.transport.set_http_response(200, self._empty_content(), headers={"Content-Type": "application/json"}):
            workspaces = await self.workspace_client.list_workspaces()
        self.assert_request_made(
            method=RequestMethod.GET,
            path=f"{BASE_PATH}/workspaces?offset=0",
            headers={"Accept": "application/json"},
        )
        self.assertEqual([], workspaces.items())

    async def test_list_workspaces_all_args(self):
        with self.transport.set_http_response(200, self._empty_content(), headers={"Content-Type": "application/json"}):
            workspaces = await self.workspace_client.list_workspaces(offset=10, limit=20)
        self.assert_request_made(
            method=RequestMethod.GET,
            path=f"{BASE_PATH}/workspaces?limit=20&offset=10",
            headers={"Accept": "application/json"},
        )
        self.assertEqual([], workspaces.items())

    async def test_delete_workspace_call(self):
        with self.transport.set_http_response(204):
            response = await self.workspace_client.delete_workspace(workspace_id=TEST_WORKSPACE_A.id)
        self.assert_request_made(
            method=RequestMethod.DELETE,
            path=f"{BASE_PATH}/workspaces/{TEST_WORKSPACE_A.id}",
        )
        self.assertIsNone(response, "Delete workspace response should be None")

    async def test_create_workspace(self):
        with self.transport.set_http_response(201, json.dumps(load_test_data("new_workspace.json"))):
            new_workspace = await self.workspace_client.create_workspace(
                name="Test Workspace",
                description="test workspace",
            )
        self.assert_request_made(
            method=RequestMethod.POST,
            path=f"{BASE_PATH}/workspaces",
            headers={
                "Accept": "application/json",
                "Content-Type": "application/json",
            },
            body={
                "bounding_box": None,
                "default_coordinate_system": "",
                "description": "test workspace",
                "labels": None,
                "name": "Test Workspace",
            },
        )
        self.assertEqual(TEST_WORKSPACE_A, new_workspace)

    async def test_update_workspace(self):
        with self.transport.set_http_response(200, json.dumps(load_test_data("new_workspace.json"))):
            updated_workspace = await self.workspace_client.update_workspace(
                workspace_id=TEST_WORKSPACE_A.id,
                name="Test Workspace",
            )
        self.assert_request_made(
            method=RequestMethod.PATCH,
            path=f"{BASE_PATH}/workspaces/{TEST_WORKSPACE_A.id}",
            headers={
                "Accept": "application/json",
                "Content-Type": "application/json",
            },
            body={
                "name": "Test Workspace",
            },
        )
        self.assertEqual(TEST_WORKSPACE_A, updated_workspace)

    async def test_assign_user_role(self) -> None:
        with self.transport.set_http_response(
            201,
            json.dumps(
                {
                    "user_id": str(USER_ID),
                    "role": "owner",
                }
            ),
        ):
            response = await self.workspace_client.assign_user_role(
                workspace_id=TEST_WORKSPACE_A.id,
                user_id=USER_ID,
                role=WorkspaceRole.owner,
            )
        self.assert_request_made(
            method=RequestMethod.POST,
            path=f"{BASE_PATH}/workspaces/{TEST_WORKSPACE_A.id}/users",
            headers={
                "Accept": "application/json",
                "Content-Type": "application/json",
            },
            body={
                "user_id": str(USER_ID),
                "role": "owner",
            },
        )
        self.assertEqual(response, UserRole(user_id=USER_ID, role=WorkspaceRole.owner))

    async def test_get_current_user_role(self) -> None:
        with self.transport.set_http_response(
            200,
            json.dumps(
                {
                    "user_id": str(USER_ID),
                    "role": "owner",
                }
            ),
        ):
            response = await self.workspace_client.get_current_user_role(workspace_id=TEST_WORKSPACE_A.id)
        self.assert_request_made(
            method=RequestMethod.GET,
            path=f"{BASE_PATH}/workspaces/{TEST_WORKSPACE_A.id}/current-user-role",
            headers={"Accept": "application/json"},
        )
        self.assertEqual(response, UserRole(user_id=USER_ID, role=WorkspaceRole.owner))

    async def test_list_user_roles(self) -> None:
        with self.transport.set_http_response(
            200,
            json.dumps(
                {
                    "results": [
                        {
                            "user_id": str(USER_ID),
                            "role": "owner",
                            "full_name": "Test User",
                            "email": "test@example.com",
                        },
                    ],
                    "links": {"self": "dummy-link.com"},
                }
            ),
        ):
            response = await self.workspace_client.list_user_roles(workspace_id=TEST_WORKSPACE_A.id)

        self.assert_request_made(
            method=RequestMethod.GET,
            path=f"{BASE_PATH}/workspaces/{TEST_WORKSPACE_A.id}/users",
            headers={"Accept": "application/json"},
        )
        self.assertEqual(
            response,
            [
                User(user_id=USER_ID, role=WorkspaceRole.owner, full_name="Test User", email="test@example.com"),
            ],
        )

    async def test_list_workspaces(self) -> None:
        content = load_test_data("list_workspaces_0.json")
        with self.transport.set_http_response(200, json.dumps(content), headers={"Content-Type": "application/json"}):
            workspaces = await self.workspace_client.list_workspaces(limit=2)
        self.assert_request_made(
            method=RequestMethod.GET,
            path=f"{BASE_PATH}/workspaces?limit=2&offset=0",
            headers={"Accept": "application/json"},
        )
        self.assertEqual(1, self.transport.request.call_count, "One requests should be made.")
        self.assertEqual([TEST_WORKSPACE_A, TEST_WORKSPACE_B], workspaces.items())

    async def test_list_all_workspaces(self) -> None:
        content_0 = load_test_data("list_workspaces_0.json")
        content_1 = load_test_data("list_workspaces_1.json")
        responses = [
            MockResponse(status_code=200, content=json.dumps(content_0), headers={"Content-Type": "application/json"}),
            MockResponse(status_code=200, content=json.dumps(content_1), headers={"Content-Type": "application/json"}),
        ]
        self.transport.request.side_effect = responses
        expected_workspaces = [TEST_WORKSPACE_A, TEST_WORKSPACE_B, TEST_WORKSPACE_C]
        actual_workspaces = await self.workspace_client.list_all_workspaces(limit=2)

        self.assertEqual(2, self.transport.request.call_count, "Two requests should be made.")
        self.assert_any_request_made(
            method=RequestMethod.GET,
            path=f"{BASE_PATH}/workspaces?limit=2&offset=0",
            headers=TestHTTPHeaderDict({"Accept": "application/json"}),
        )
        self.assert_any_request_made(
            method=RequestMethod.GET,
            path=f"{BASE_PATH}/workspaces?limit=2&offset=2",
            headers=TestHTTPHeaderDict({"Accept": "application/json"}),
        )
        self.assertEqual(expected_workspaces, actual_workspaces)

    async def test_list_workspaces_sorted_by_display_name(self) -> None:
        content_0 = load_test_data("list_workspaces_0.json")
        content_1 = load_test_data("list_workspaces_1.json")

        # Shuffle the workspaces in the response content.
        results_0: list = content_0["results"]
        results_1: list = content_1["results"]
        results_0.insert(2, results_1.pop(0))
        results_1.insert(0, results_0.pop(0))

        responses = [
            MockResponse(status_code=200, content=json.dumps(content_0), headers={"Content-Type": "application/json"}),
            MockResponse(status_code=200, content=json.dumps(content_1), headers={"Content-Type": "application/json"}),
        ]
        self.transport.request.side_effect = responses
        expected_workspaces = [TEST_WORKSPACE_A, TEST_WORKSPACE_B, TEST_WORKSPACE_C]
        actual_workspaces = await self.workspace_client.list_all_workspaces(limit=2)

        self.assertEqual(2, self.transport.request.call_count, "Two requests should be made.")
        self.assert_any_request_made(
            method=RequestMethod.GET,
            path=f"{BASE_PATH}/workspaces?limit=2&offset=0",
            headers=TestHTTPHeaderDict({"Accept": "application/json"}),
        )
        self.assert_any_request_made(
            method=RequestMethod.GET,
            path=f"{BASE_PATH}/workspaces?limit=2&offset=2",
            headers=TestHTTPHeaderDict({"Accept": "application/json"}),
        )
        self.assertEqual(expected_workspaces, actual_workspaces)

    async def test_list_workspaces_summary_default_args(self):
        with self.transport.set_http_response(200, self._empty_content(), headers={"Content-Type": "application/json"}):
            workspaces = await self.workspace_client.list_workspaces_summary()
        self.assert_request_made(
            method=RequestMethod.GET,
            path=f"{BASE_PATH}/workspaces/summary",
            headers={"Accept": "application/json"},
        )
        self.assertEqual([], workspaces.items())

    async def test_list_workspaces_summary_all_args(self) -> None:
        for order_by in [
            {WorkspaceOrderByEnum.name: OrderByOperatorEnum.asc},
            {"name": "asc"},
        ]:
            with self.transport.set_http_response(
                200, self._empty_content(), headers={"Content-Type": "application/json"}
            ):
                workspaces = await self.workspace_client.list_workspaces_summary(
                    offset=10,
                    limit=20,
                    order_by=order_by,
                    filter_created_by=USER_ID,
                    created_at=str(utc_datetime(2020, 1, 1)),
                    updated_at=str(utc_datetime(2020, 1, 1)),
                    name="Test Workspace A",
                    deleted=False,
                    filter_user_id=USER_ID,
                )
            self.assert_request_made(
                method=RequestMethod.GET,
                path=f"{BASE_PATH}/workspaces/summary?"
                f"limit=20&offset=10&order_by=asc%3Aname&filter%5Bcreated_by%5D=00000000-0000-0000-0000-000000000002&"
                f"created_at=2020-01-01+00%3A00%3A00%2B00%3A00&updated_at=2020-01-01+00%3A00%3A00%2B00%3A00&"
                f"name=Test+Workspace+A&deleted=False&filter%5Buser_id%5D=00000000-0000-0000-0000-000000000002",
                headers={"Accept": "application/json"},
            )
            self.assertEqual([], workspaces.items())

    async def test_list_workspaces_summary(self) -> None:
        content = load_test_data("list_workspaces_summary.json")
        with self.transport.set_http_response(200, json.dumps(content), headers={"Content-Type": "application/json"}):
            workspaces = await self.workspace_client.list_workspaces_summary()
        self.assert_request_made(
            method=RequestMethod.GET,
            path=f"{BASE_PATH}/workspaces/summary",
            headers={"Accept": "application/json"},
        )
        self.assertEqual(1, self.transport.request.call_count, "One requests should be made.")
        self.assertEqual([TEST_BASIC_WORKSPACE_A, TEST_BASIC_WORKSPACE_B, TEST_BASIC_WORKSPACE_C], workspaces.items())

    async def test_paginated_list_workspaces_summary(self) -> None:
        content = load_test_data("list_workspaces_summary_paginated_0.json")
        content_2 = load_test_data("list_workspaces_summary_paginated_1.json")
        with self.transport.set_http_response(200, json.dumps(content), headers={"Content-Type": "application/json"}):
            workspaces_page_1 = await self.workspace_client.list_workspaces_summary(limit=2, offset=0)

        with self.transport.set_http_response(200, json.dumps(content_2), headers={"Content-Type": "application/json"}):
            workspaces_page_2 = await self.workspace_client.list_workspaces_summary(limit=2, offset=2)

        self.assert_any_request_made(
            method=RequestMethod.GET,
            path=f"{BASE_PATH}/workspaces/summary?limit=2&offset=0",
            headers={"Accept": "application/json"},
        )
        self.assert_any_request_made(
            method=RequestMethod.GET,
            path=f"{BASE_PATH}/workspaces/summary?limit=2&offset=2",
            headers={"Accept": "application/json"},
        )
        self.assertEqual(2, self.transport.request.call_count, "Two requests should be made.")
        self.assertEqual([TEST_BASIC_WORKSPACE_A, TEST_BASIC_WORKSPACE_B], workspaces_page_1.items())
        self.assertEqual([TEST_BASIC_WORKSPACE_C], workspaces_page_2.items())



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import datetime
import json
from collections.abc import Iterator, Mapping
from contextlib import contextmanager
from enum import Enum
from typing import Any
from unittest import mock
from uuid import UUID

from parameterized import parameterized
from pydantic import BaseModel

from evo.common import APIConnector, EmptyResponse, HTTPHeaderDict, HTTPResponse, RequestMethod
from evo.common.exceptions import (
    BadRequestException,
    ClientValueError,
    CustomTypedError,
    DefaultTypedError,
    EvoAPIException,
    ForbiddenException,
    GoneException,
    NotFoundException,
    UnauthorizedException,
    UnknownResponseError,
)
from evo.common.test_tools import MockResponse, TestWithConnector, utc_datetime, utc_time


class SampleEnum(Enum):
    STRING_VALUE = "string value"
    INTEGER_VALUE = 1
    FLOAT_VALUE = 1.1
    BYTES_VALUE = b"\xab\xcd"


class SamplePydanticModel(BaseModel):
    str_value: str = "string value"
    int_value: int = 1
    float_value: float = 1.1
    enum_value: SampleEnum = SampleEnum.STRING_VALUE
    datetime_value: datetime.datetime = utc_datetime(2000, 1, 2, 3, 4, 5)
    uuid_value: UUID = UUID(int=1)


class _ResponseType200(BaseModel):
    value: str


class _ResponseType201(BaseModel):
    str_value: str
    int_value: int
    float_value: float
    bool_value: bool
    nested_value: _ResponseType200


class _ResponseType203(_ResponseType200):
    other_value: str


class TestAPIConnector(TestWithConnector):
    def setUp(self) -> None:
        super().setUp()
        self.transport.request.return_value = MockResponse(status_code=200)

    async def test_open(self) -> None:
        """Test that the connector opens the transport when opened."""
        self.transport.open.assert_not_called()
        await self.connector.open()
        self.transport.open.assert_called_once()

    async def test_close(self) -> None:
        """Test that the connector closes the transport when closed."""
        self.transport.close.assert_not_called()
        await self.connector.close()
        self.transport.close.assert_called_once()

    async def test_context_manager(self) -> None:
        """Test that the connector opens and closes the transport when used as a context manager."""
        self.transport.open.assert_not_called()
        self.transport.close.assert_not_called()
        async with self.connector:
            self.transport.open.assert_called_once()
            self.transport.close.assert_not_called()
        self.transport.open.assert_called_once()
        self.transport.close.assert_called_once()

    async def test_call_api_manages_transport(self) -> None:
        """Test that the connector opens and closes the transport when making an API call."""
        self.transport.open.assert_not_called()
        self.transport.close.assert_not_called()
        await self.connector.call_api(method=RequestMethod.GET, resource_path="", response_types_map={"200": str})
        self.transport.open.assert_called_once()
        self.transport.close.assert_called_once()

    async def test_call_api_enters_context_once(self) -> None:
        """Test that the connector uses the same context for retrying an API call."""
        self.authorizer.set_next_access_token("<new-access-token>")

        self.transport.request.side_effect = [
            MockResponse(status_code=401, content="Unauthorized", reason="Unauthorized"),
            MockResponse(status_code=200, content="success"),
        ]
        self.transport.open.assert_not_called()
        self.transport.close.assert_not_called()
        await self.connector.call_api(
            method=RequestMethod.GET,
            resource_path="",
            response_types_map={"200": str},
        )
        self.authorizer.refresh_token.assert_called_once()
        self.transport.open.assert_called_once()
        self.transport.close.assert_called_once()

    @parameterized.expand([(method,) for method in RequestMethod])
    async def test_request_method(self, method: RequestMethod) -> None:
        with self.transport.set_http_response(status_code=200, content="success"):
            result = await self.connector.call_api(method=method, resource_path="", response_types_map={"200": str})
        self.assertEqual("success", result)
        self.assert_request_made(method=method)

    @parameterized.expand(
        [
            (  # Single path param.
                "/service/{single-param}",
                {"single-param": "str-value"},
                "/service/str-value",
            ),
            (  # Multiple path params.
                "/service/{org}/{workspace}/",
                {"org": "org-id", "workspace": UUID(int=1)},
                "/service/org-id/00000000-0000-0000-0000-000000000001/",
            ),
        ]
    )
    async def test_path_params(self, path: str, params: dict[str, str | object], expected_path: str) -> None:
        await self.connector.call_api(
            method=RequestMethod.GET,
            resource_path=path,
            path_params=params,
            response_types_map={"200": bytes},
        )
        self.assert_request_made(method=RequestMethod.GET, path=expected_path)

    @parameterized.expand(
        [
            (
                "Single query param",
                {"param": "value"},
                "?param=value",
            ),
            (
                "Multiple query params in alphanumeric order",
                {"param1": "value1", "param2": "value2", "param3": "value3"},
                "?param1=value1&param2=value2&param3=value3",
            ),
            (
                "Multiple query params in random order",
                {"param3": "value3", "param1": "value1", "param2": "value2"},
                "?param3=value3&param1=value1&param2=value2",
            ),
            (
                "Multi-value query param in alphanumeric order",
                {"param": ["value1", "value2", "value3"]},
                "?param=value1&param=value2&param=value3",
                "multi",
            ),
            (
                "Multi-value query param in random order",
                {"param": ["value3", "value1", "value2"]},
                "?param=value3&param=value1&param=value2",
                "multi",
            ),
            (
                "SSV-value query param in alphanumeric order",
                {"param": ["value1", "value2", "value3"]},
                "?param=value1+value2+value3",
                "ssv",
            ),
            (
                "SSV-value query param in random order",
                {"param": ["value3", "value1", "value2"]},
                "?param=value3+value1+value2",
                "ssv",
            ),
            (
                "Pipe-separated value query param in alphanumeric order",
                {"param": ["value1", "value2", "value3"]},
                "?param=value1%7Cvalue2%7Cvalue3",
                "pipes",
            ),
            (
                "Pipe-separated value query param in random order",
                {"param": ["value3", "value1", "value2"]},
                "?param=value3%7Cvalue1%7Cvalue2",
                "pipes",
            ),
            (
                "CSV-value query param in alphanumeric order",
                {"param": ["value1", "value2", "value3"]},
                "?param=value1%2Cvalue2%2Cvalue3",
                "csv",
            ),
            (
                "CSV-value query param in random order",
                {"param": ["value3", "value1", "value2"]},
                "?param=value3%2Cvalue1%2Cvalue2",
                "csv",
            ),
        ]
    )
    async def test_query_params(
        self,
        _name: str,
        params: dict[str, Any],
        expected_path: str,
        collection_format: str | None = None,
    ) -> None:
        collection_formats = {param: collection_format for param in params.keys()} if collection_format else None
        await self.connector.call_api(
            method=RequestMethod.GET,
            resource_path="",
            query_params=params,
            collection_formats=collection_formats,
            response_types_map={"200": bytes},
        )
        self.assert_request_made(method=RequestMethod.GET, path=expected_path)

    @parameterized.expand(
        [
            (
                "Single header param",
                {"param": "value"},
                {"param": "value"},
            ),
            (
                "Multiple header params",
                {"param1": "value1", "param2": "value2", "param3": "value3"},
                {"param1": "value1", "param2": "value2", "param3": "value3"},
            ),
            (
                "SSV-value header param in alphanumeric order",
                {"param": ["value1", "value2", "value3"]},
                {"param": "value1 value2 value3"},
                "ssv",
            ),
            (
                "SSV-value header param in random order",
                {"param": ["value3", "value1", "value2"]},
                {"param": "value3 value1 value2"},
                "ssv",
            ),
            (
                "Pipe-separated value header param in alphanumeric order",
                {"param": ["value1", "value2", "value3"]},
                {"param": "value1|value2|value3"},
                "pipes",
            ),
            (
                "Pipe-separated value header param in random order",
                {"param": ["value3", "value1", "value2"]},
                {"param": "value3|value1|value2"},
                "pipes",
            ),
            (
                "CSV-value header param in alphanumeric order",
                {"param": ["value1", "value2", "value3"]},
                {"param": "value1,value2,value3"},
                "csv",
            ),
            (
                "CSV-value header param in random order",
                {"param": ["value3", "value1", "value2"]},
                {"param": "value3,value1,value2"},
                "csv",
            ),
            (
                "Single header param as mapping",
                HTTPHeaderDict(param="value"),
                {"param": "value"},
            ),
        ]
    )
    async def test_header_params(
        self,
        _name: str,
        params: Mapping[str, Any],
        expected_headers: Mapping[str, str],
        collection_format: str | None = None,
    ) -> None:
        collection_formats = {param: collection_format for param in params.keys()} if collection_format else None
        await self.connector.call_api(
            method=RequestMethod.GET,
            resource_path="",
            header_params=params,
            collection_formats=collection_formats,
            response_types_map={"200": bytes},
        )
        self.assert_request_made(method=RequestMethod.GET, headers=expected_headers)

    @parameterized.expand(
        [
            (
                "Single header param",
                {"param": "value"},
            ),
            (
                "Multiple header params",
                {"param1": "value1", "param2": "value2", "param3": "value3"},
            ),
        ]
    )
    async def test_additional_headers_in_connector(
        self,
        _name: str,
        headers: Mapping[str, str],
    ) -> None:
        self.connector._additional_headers = HTTPHeaderDict(headers)
        await self.connector.call_api(
            method=RequestMethod.GET,
            resource_path="",
            response_types_map={"200": bytes},
        )
        self.assert_request_made(method=RequestMethod.GET, headers=HTTPHeaderDict(headers))

    async def test_header_params_multi_value_raises_error(self) -> None:
        with self.assertRaises(RuntimeError):
            await self.connector.call_api(
                method=RequestMethod.GET,
                resource_path="",
                header_params={"param": ["value1", "value2", "value3"]},
                collection_formats={"param": "multi"},
                response_types_map={"200": bytes},
            )
            self.transport.request.assert_not_called()

    @parameterized.expand(
        [
            (
                "Single post param",
                {"param": "value"},
                [("param", "value")],
            ),
            (
                "Multiple post params in alphanumeric order",
                {"param1": "value1", "param2": "value2", "param3": "value3"},
                [("param1", "value1"), ("param2", "value2"), ("param3", "value3")],
            ),
            (
                "Multiple post params in random order",
                {"param3": "value3", "param1": "value1", "param2": "value2"},
                [("param3", "value3"), ("param1", "value1"), ("param2", "value2")],
            ),
            (
                "Multi-value post param in alphanumeric order",
                {"param": ["value1", "value2", "value3"]},
                [("param", "value1"), ("param", "value2"), ("param", "value3")],
                "multi",
            ),
            (
                "Multi-value post param in random order",
                {"param": ["value3", "value1", "value2"]},
                [("param", "value3"), ("param", "value1"), ("param", "value2")],
                "multi",
            ),
            (
                "SSV-value post param in alphanumeric order",
                {"param": ["value1", "value2", "value3"]},
                [("param", "value1 value2 value3")],
                "ssv",
            ),
            (
                "SSV-value post param in random order",
                {"param": ["value3", "value1", "value2"]},
                [("param", "value3 value1 value2")],
                "ssv",
            ),
            (
                "Pipe-separated value post param in alphanumeric order",
                {"param": ["value1", "value2", "value3"]},
                [("param", "value1|value2|value3")],
                "pipes",
            ),
            (
                "Pipe-separated value post param in random order",
                {"param": ["value3", "value1", "value2"]},
                [("param", "value3|value1|value2")],
                "pipes",
            ),
            (
                "CSV-value post param in alphanumeric order",
                {"param": ["value1", "value2", "value3"]},
                [("param", "value1,value2,value3")],
                "csv",
            ),
            (
                "CSV-value post param in random order",
                {"param": ["value3", "value1", "value2"]},
                [("param", "value3,value1,value2")],
                "csv",
            ),
        ]
    )
    async def test_post_params(
        self,
        _name: str,
        params: dict[str, Any],
        expected_post_params: list[tuple[str, str]],
        collection_format: str | None = None,
    ) -> None:
        collection_formats = {param: collection_format for param in params.keys()} if collection_format else None
        await self.connector.call_api(
            method=RequestMethod.GET,
            resource_path="",
            post_params=params,
            collection_formats=collection_formats,
            response_types_map={"200": bytes},
        )
        self.assert_request_made(
            method=RequestMethod.GET,
            post_params=expected_post_params,
        )

    async def _parse_response(self, expected_type: Any, response: MockResponse | None = None) -> Any:
        if response is not None:
            self.transport.request.return_value = response
        result = await self.connector.call_api(
            method=RequestMethod.GET,
            resource_path="",
            response_types_map={"200": expected_type},
        )
        self.assert_request_made(method=RequestMethod.GET)
        return result

    @parameterized.expand(
        [
            (
                "empty response",
                "",
                EmptyResponse(status=200, headers=HTTPHeaderDict({"Sample-Header": "sample value"})),
                EmptyResponse,
            ),
            (
                "string",
                "a simple string",
                "a simple string",
                str,
            ),
            (
                "integer",
                "1",
                1,
                int,
            ),
            (
                "float",
                "2.2",
                2.2,
                float,
            ),
            (
                "boolean (true)",
                "true",
                True,
                bool,
            ),
            (
                "boolean (false)",
                "false",
                False,
                bool,
            ),
            (
                "bytes",
                "abcd",
                b"\x61\x62\x63\x64",
                bytes,
            ),
            (
                "time",
                "13:05:04.321+00:00",
                utc_time(hour=13, minute=5, second=4, microsecond=321000),
                datetime.time,
            ),
            (
                "date",
                "2023-05-23",
                datetime.date(year=2023, month=5, day=23),
                datetime.date,
            ),
            (
                "datetime",
                "2023-05-23T13:05:04.321+00:00",
                utc_datetime(
                    year=2023,
                    month=5,
                    day=23,
                    hour=13,
                    minute=5,
                    second=4,
                    microsecond=321000,
                ),
                datetime.datetime,
            ),
            (
                "dict",
                """
                {
                    "string": "a string",
                    "integer": 1,
                    "float": 2.2,
                    "boolean": true,
                    "time": "13:00:00",
                    "date": "2023-05-23",
                    "datetime": "2023-05-23T13:00:00+00:00"
                }
                """,
                {
                    "string": "a string",
                    "integer": 1,
                    "float": 2.2,
                    "boolean": True,
                    "time": "13:00:00",
                    "date": "2023-05-23",
                    "datetime": "2023-05-23T13:00:00+00:00",
                },
                dict,
            ),
            (
                "Basic pydantic model",
                """
                {"value": "basic value"}
                """,
                _ResponseType200(value="basic value"),
                _ResponseType200,
            ),
            (
                "Complex pydantic model",
                """
                {
                    "str_value": "a string",
                    "int_value": 1,
                    "float_value": 2.2,
                    "bool_value": true,
                    "nested_value": {
                        "value": "another string"
                    }
                }
                """,
                _ResponseType201(
                    str_value="a string",
                    int_value=1,
                    float_value=2.2,
                    bool_value=True,
                    nested_value=_ResponseType200(value="another string"),
                ),
                _ResponseType201,
            ),
            (
                "No response type expecting json",
                '{"key": "value"}',
                {"key": "value"},
                None,
            ),
            (
                "No response type expecting string",
                "a simple string",
                "a simple string",
                None,
            ),
        ]
    )
    async def test_response_types(
        self,
        _name: str,
        response_content: str,
        expected_object: Any,
        expected_type: type,
    ):
        mock_response = MockResponse(
            status_code=200, content=response_content, headers={"sample-header": "sample value"}
        )
        actual_obj = await self._parse_response(
            expected_type=expected_type,
            response=mock_response,
        )
        self.assertEqual(expected_object, actual_obj)
        if isinstance(expected_object, EmptyResponse):
            mock_response.getheaders.assert_called_once()

    @parameterized.expand(
        [
            (
                "list of strings",
                '["one", "two", "three", "four"]',
                list[str],
                ["one", "two", "three", "four"],
            ),
            (
                "list of integers",
                "[1, 2, 3, 4]",
                list[int],
                [1, 2, 3, 4],
            ),
            (
                "list of floats",
                "[1.1, 2.2, 3.3, 4.4]",
                list[float],
                [1.1, 2.2, 3.3, 4.4],
            ),
            (
                "list of booleans",
                "[true, true, false, false, true]",
                list[bool],
                [True, True, False, False, True],
            ),
        ]
    )
    async def test_generic_response_types(
        self,
        _name: str,
        response_content: str,
        expected_type: Any,
        expected_value: Any,
    ) -> None:
        actual_value = await self._parse_response(
            expected_type=expected_type,
            response=MockResponse(status_code=200, content=response_content),
        )
        self.assertEqual(expected_value, actual_value)

    @parameterized.expand(
        [
            ("Normal response", 200),
            ("Expected generic error response does not raise", 404),
            # The following happens to be the status code used for failed service health checks.
            ("Expected error response does not raise", 503),
        ]
    )
    async def test_raw_response(self, name: str, status_code: int) -> None:
        """Test that the connector always returns the expected response for mapped status codes."""
        with self.transport.set_http_response(status_code=status_code, content=name):
            response = await self.connector.call_api(
                method=RequestMethod.GET,
                resource_path="",
                response_types_map={str(status_code): HTTPResponse},
            )
        self.assert_request_made(method=RequestMethod.GET)
        self.assertIsInstance(response, HTTPResponse)
        self.assertEqual(status_code, response.status)
        self.assertEqual(name, response.data.decode())

    async def test_empty_response_error(self) -> None:
        with (
            self.transport.set_http_response(
                status_code=204, content="some content", headers={"Sample-Header": "sample header value"}
            ),
            self.assertRaises(ClientValueError) as cm,
        ):
            await self.connector.call_api(
                method=RequestMethod.GET,
                resource_path="",
                response_types_map={"204": EmptyResponse},
            )
        expected_message = "Unexpected content with '204' status code"
        actual_message = str(cm.exception)
        self.assertEqual(expected_message, actual_message)

    @parameterized.expand(
        [
            (
                "Invalid response (200)",
                MockResponse(
                    status_code=200,
                    content='{"other_value": "other value"}',
                    reason="OK",
                ),
                ClientValueError,
                "Could not deserialize result: 1 validation error for _ResponseType200",
            ),
            (
                "Unknown response type (203)",
                MockResponse(
                    status_code=203,
                    content="some content",
                    reason="OK",
                ),
                UnknownResponseError,
                "(203) OK\nsome content",
            ),
            (
                "Unknown response type (203) with body",
                MockResponse(
                    status_code=203,
                    content=json.dumps(
                        {
                            "type": "service data",
                            "title": "Service Data Format",
                            "detail": "some data model",
                        }
                    ),
                    reason="OK",
                ),
                UnknownResponseError,
                "(203) OK\n{'type': 'service data', 'title': 'Service Data Format', 'detail': 'some data model'}",
            ),
            (
                "Bad request (400)",
                MockResponse(
                    status_code=400,
                    content=json.dumps({"title": "Bad request title.", "detail": "(optional) additional detail."}),
                    reason="Bad Request",
                ),
                BadRequestException,
                "Error: (400) Bad Request"
                "\nType: about:blank"
                "\nTitle: Bad request title."
                "\nDetail: (optional) additional detail.",
            ),
            (
                "Bad request (401)",
                MockResponse(
                    status_code=401,
                    content=json.dumps({"title": "Unauthorized title."}),
                    reason="Unauthorized",
                ),
                UnauthorizedException,
                "Error: (401) Unauthorized\nType: about:blank\nTitle: Unauthorized title.",
            ),
            (
                "Access denied (403)",
                MockResponse(
                    status_code=403,
                    content=json.dumps({"title": "Access denied title."}),
                    reason="Access Denied",
                ),
                ForbiddenException,
                "Error: (403) Access Denied\nType: about:blank\nTitle: Access denied title.",
            ),
            (
                "Not found (404)",
                MockResponse(
                    status_code=404,
                    content=json.dumps({"title": "Not found title."}),
                    reason="Not Found",
                ),
                NotFoundException,
                "Error: (404) Not Found\nType: about:blank\nTitle: Not found title.",
            ),
            (
                "Conflict (409)",
                MockResponse(
                    status_code=409,
                    content=json.dumps({"title": "Conflict title."}),
                    reason="Conflict",
                ),
                DefaultTypedError,
                "Error: (409) Conflict\nType: about:blank\nTitle: Conflict title.",
            ),
            (
                "Gone (410)",
                MockResponse(
                    status_code=410,
                    content=json.dumps({"title": "Gone title."}),
                    reason="Gone",
                ),
                GoneException,
                "Error: (410) Gone\nType: about:blank\nTitle: Gone title.",
            ),
            (
                "Internal server error (500)",
                MockResponse(
                    status_code=500,
                    content=json.dumps({"title": "Internal server error title."}),
                    reason="Internal Server Error",
                ),
                DefaultTypedError,
                "Error: (500) Internal Server Error\nType: about:blank\nTitle: Internal server error title.",
            ),
            (
                "Bad gateway (502)",
                MockResponse(
                    status_code=502,
                    content=json.dumps({"title": "Bad gateway title."}),
                    reason="Bad Gateway",
                ),
                DefaultTypedError,
                "Error: (502) Bad Gateway\nType: about:blank\nTitle: Bad gateway title.",
            ),
            (
                "Internal server error (500) with specific type id",
                MockResponse(
                    status_code=500,
                    content=json.dumps(
                        {
                            "title": "Internal server error title.",
                            "type": "https://specific.unittest.test/errors/internal-server-error",
                        }
                    ),
                    reason="Internal Server Error",
                ),
                DefaultTypedError,
                "Error: (500) Internal Server Error"
                "\nType: https://specific.unittest.test/errors/internal-server-error"
                "\nTitle: Internal server error title.",
            ),
            (
                "Internal server error (500) with other type id",
                MockResponse(
                    status_code=500,
                    content=json.dumps(
                        {
                            "title": "Internal server error title.",
                            "type": "https://other.unittest.test/errors/other/internal-server-error",
                        }
                    ),
                    reason="Internal Server Error",
                ),
                DefaultTypedError,
                "Error: (500) Internal Server Error"
                "\nType: https://other.unittest.test/errors/other/internal-server-error"
                "\nTitle: Internal server error title.",
            ),
            (
                "Unauthorized (401) with specific type id",
                MockResponse(
                    status_code=401,
                    content=json.dumps(
                        {"title": "Unauthorized title.", "type": "https://specific.unittest.test/errors/unauthorized"}
                    ),
                    reason="Unauthorized",
                ),
                UnauthorizedException,
                "Error: (401) Unauthorized"
                "\nType: https://specific.unittest.test/errors/unauthorized"
                "\nTitle: Unauthorized title.",
            ),
            (
                "Unauthorized (401) with other type id",
                MockResponse(
                    status_code=401,
                    content=json.dumps(
                        {
                            "title": "Unauthorized title.",
                            "type": "https://other.unittest.test/errors/other/unauthorized",
                        }
                    ),
                    reason="Unauthorized",
                ),
                UnauthorizedException,
                "Error: (401) Unauthorized"
                "\nType: https://other.unittest.test/errors/other/unauthorized"
                "\nTitle: Unauthorized title.",
            ),
        ]
    )
    async def test_error_response(
        self,
        _name: str,
        response: MockResponse,
        exc_type: type[EvoAPIException],
        expected_message: str,
    ) -> None:
        with self.assertRaises(exc_type) as cm:
            await self._parse_response(expected_type=_ResponseType200, response=response)
        actual_exc_type = type(cm.exception)
        self.assertIs(exc_type, actual_exc_type)
        actual_message = str(cm.exception)
        self.assertIn(expected_message, actual_message)

    @parameterized.expand(
        [
            # If obj is None return None.
            ("no value", None, None),
            # If obj is an Enum sanitize the value.
            ("enum string value", SampleEnum.STRING_VALUE, "string value"),
            ("enum integer value", SampleEnum.INTEGER_VALUE, 1),
            ("enum float value", SampleEnum.FLOAT_VALUE, 1.1),
            ("enum bytes value", SampleEnum.BYTES_VALUE, b"\xab\xcd"),
            # If obj is a primitive return directly.
            ("string", "string value", "string value"),
            ("integer", 1, 1),
            ("float", 1.1, 1.1),
            ("boolean", True, True),
            ("bytes", b"\xab\xcd", b"\xab\xcd"),
            ("boolean", True, True),
            # If obj is a date or datetime convert to string in iso8601 format.
            ("datetime", True, True),
            # If obj is a UUID convert to string.
            ("uuid", UUID(int=1), "00000000-0000-0000-0000-000000000001"),
            # If obj is a list or tuple, sanitize each element.
            (
                "list",
                [
                    None,
                    SampleEnum.STRING_VALUE,
                    "string value",
                    1,
                    1.1,
                    True,
                    b"\xab\xcd",
                    utc_datetime(2000, 1, 2, 3, 4, 5),
                ],
                [
                    None,
                    "string value",
                    "string value",
                    1,
                    1.1,
                    True,
                    b"\xab\xcd",
                    "2000-01-02T03:04:05+00:00",
                ],
            ),
            (
                "tuple",
                (
                    None,
                    SampleEnum.STRING_VALUE,
                    "string value",
                    1,
                    1.1,
                    True,
                    b"\xab\xcd",
                    utc_datetime(2000, 1, 2, 3, 4, 5),
                ),
                (
                    None,
                    "string value",
                    "string value",
                    1,
                    1.1,
                    True,
                    b"\xab\xcd",
                    "2000-01-02T03:04:05+00:00",
                ),
            ),
            # If obj is a dict, sanitize the dict.
            (
                "dict",
                {
                    "none": None,
                    "enum": SampleEnum.STRING_VALUE,
                    "string": "string value",
                    "int": 1,
                    "float": 1.1,
                    "bool": True,
                    "bytes": b"\xab\xcd",
                    "datetime": utc_datetime(2000, 1, 2, 3, 4, 5),
                },
                {
                    "none": None,
                    "enum": "string value",
                    "string": "string value",
                    "int": 1,
                    "float": 1.1,
                    "bool": True,
                    "bytes": b"\xab\xcd",
                    "datetime": "2000-01-02T03:04:05+00:00",
                },
            ),
            # If obj is an API model, convert to dict.
            (
                "api model",
                SamplePydanticModel(),
                json.loads(SamplePydanticModel().model_dump_json(by_alias=True, exclude_unset=True)),
            ),
        ]
    )
    def test_sanitization(self, _name: str, input_value: Any, expected_value: Any) -> None:
        expected_type = type(expected_value)
        actual_value = APIConnector._sanitize_for_serialization(input_value)
        self.assertIsInstance(actual_value, expected_type)
        self.assertEqual(actual_value, expected_value)

    async def test_refresh_on_auth_error(self) -> None:
        """Test that the connector refreshes the access token and retries the request on an auth error."""
        old_headers = self.authorizer.default_headers.copy()
        self.authorizer.set_next_access_token("<new-access-token>")

        self.transport.request.side_effect = [
            MockResponse(status_code=401, content="Unauthorized", reason="Unauthorized"),
            MockResponse(status_code=200, content="success"),
        ]
        response = await self.connector.call_api(
            method=RequestMethod.GET,
            resource_path="",
            response_types_map={"200": str},
        )
        self.assertEqual("success", response)
        self.authorizer.refresh_token.assert_called_once()
        self.transport.assert_any_request_made(method=RequestMethod.GET, headers=old_headers)

        new_headers = self.authorizer.default_headers.copy()
        self.transport.assert_request_made(method=RequestMethod.GET, headers=new_headers)

    async def test_refresh_on_auth_error_fails(self) -> None:
        """Test that the connector raises an exception and does not retry the request if refreshing the token fails."""
        self.transport.request.side_effect = [
            MockResponse(status_code=401, content="Unauthorized", reason="Unauthorized"),
            MockResponse(status_code=200, content="success"),
        ]
        with self.assertRaises(UnauthorizedException):
            await self.connector.call_api(
                method=RequestMethod.GET,
                resource_path="",
                response_types_map={"200": str},
            )
        self.authorizer.refresh_token.assert_called_once()
        self.transport.request.assert_called_once()
        self.transport.assert_any_request_made(method=RequestMethod.GET, headers=self.authorizer.default_headers.copy())

    @contextmanager
    def __temp_register_custom_error(self) -> Iterator[None]:
        with mock.patch.object(CustomTypedError, "_CustomTypedError__CONCRETE_TYPES", dict()):
            yield

    async def test_custom_typed_error_mapped(self) -> None:
        mock_response = MockResponse(
            status_code=422,
            content=json.dumps(
                {
                    "type": "https://other.unittest.test/errors/other/validation",
                    "title": "Validation Error",
                }
            ),
        )

        with self.__temp_register_custom_error():

            class CustomValidationError(CustomTypedError):
                TYPE_ID = "errors/other/validation"

            with self.assertRaises(CustomValidationError):
                await self._parse_response(expected_type=_ResponseType200, response=mock_response)

    async def test_typed_error_fallback(self) -> None:
        mock_response = MockResponse(
            status_code=422,
            content=json.dumps(
                {
                    "type": "https://other.unittest.test/errors/other/validation",
                    "title": "Validation Error",
                }
            ),
        )

        with self.__temp_register_custom_error():

            class NotTheSameCustomError(CustomTypedError):
                TYPE_ID = "/some/other/error"

            with self.assertRaises(DefaultTypedError):
                await self._parse_response(expected_type=_ResponseType200, response=mock_response)

    def test_duplicate_custom_typed_error(self) -> None:
        def create_duplicate_custom_error_handle() -> None:
            class AnTypedError(CustomTypedError):
                TYPE_ID = "/error-source"

            class ADuplicateTypedError(CustomTypedError):
                TYPE_ID = "/error-source"

        with self.__temp_register_custom_error():
            with self.assertRaises(ValueError):
                create_duplicate_custom_error_handle()

    def test_invalid_type_id_for_custom_typed_error_handle(self) -> None:
        def create_invalid_type_id() -> None:
            class AnInvalidTypedError(CustomTypedError):
                TYPE_ID = 1

        with self.__temp_register_custom_error():
            with self.assertRaises(ValueError):
                create_invalid_type_id()

    def test_nonetype_allowed_for_abstract_typed_error_handle(self) -> None:
        with self.__temp_register_custom_error():

            class AValidBaseTypedError(CustomTypedError):
                TYPE_ID = None

            class AValidTypedError(AValidBaseTypedError):
                TYPE_ID = "/some/path/to/error"

    def test_type_id_must_be_set_for_custom_typed_error_handle(self) -> None:
        with self.__temp_register_custom_error():
            with self.assertRaises(ValueError):

                class NoTypeIDTypedError(CustomTypedError):
                    pass



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import shutil
from pathlib import Path
from uuid import uuid5

from evo.common.exceptions import StorageFileExistsError, StorageFileNotFoundError
from evo.common.test_tools import TestWithStorage
from evo.common.utils import Cache


class TestCache(TestWithStorage):
    def test_init_with_path(self) -> None:
        """Test setting the cache location with a Path object."""
        new_location = self.CACHE_DIR / "some subdir"
        shutil.rmtree(new_location, ignore_errors=True)
        self.assertFalse(new_location.exists(), "The cache directory should not exist.")

        manager = Cache(new_location, mkdir=True)
        self.assertTrue(new_location.is_dir(), "The cache directory should exist.")
        self.assertEqual(
            new_location,
            manager.get_location(self.environment, "some scope").parent,
            "The cache location should be set correctly.",
        )

    def test_init_with_str(self) -> None:
        """Test setting the cache location with a string."""
        new_location = self.CACHE_DIR / "some other subdir"
        shutil.rmtree(new_location, ignore_errors=True)
        self.assertFalse(new_location.exists(), "The cache directory should not exist.")

        manager = Cache(str(new_location), mkdir=True)
        self.assertTrue(new_location.is_dir(), "The cache directory should exist.")
        self.assertEqual(
            new_location,
            manager.get_location(self.environment, "some other scope").parent,
            "The cache location should be set correctly.",
        )

    def test_init_with_other_types_raise_typeerror(self) -> None:
        """Test setting the cache location with a string."""
        for invalid in (1, 1.0, b"some bytes", False, object(), dict(), list(), tuple(), set()):
            with self.subTest(invalid=invalid):
                with self.assertRaises(TypeError):
                    Cache(invalid)  # type: ignore

    def test_init_with_nonexistent(self) -> None:
        """Test setting the cache location to a non-existent directory when mkdir=False should raise an error."""
        new_location = self.CACHE_DIR / "nonexistent"
        shutil.rmtree(new_location, ignore_errors=True)
        self.assertFalse(new_location.exists(), "The cache directory should not exist.")

        with self.assertRaises(StorageFileNotFoundError):
            Cache(new_location)

    def test_init_with_file_exists(self) -> None:
        """Test setting the cache location to an existing file should raise an error."""
        new_location = self.CACHE_DIR / "some file"
        new_location.touch()
        self.assertTrue(new_location.is_file(), "The cache directory should be a file.")

        with self.assertRaises(StorageFileExistsError):
            Cache(new_location)

    def test_root(self) -> None:
        """Test getting the cache root."""
        self.assertEqual(self.CACHE_DIR, self.cache.root)

    def test_get_cache(self) -> None:
        """Test getting the cache directory for a given scope."""
        scope = "some random scope"
        expected_cache_name = str(uuid5(self.environment.workspace_id, scope))
        cache_dir = self.cache.get_location(self.environment, scope)
        self.assertIsInstance(cache_dir, Path, "The cache directory should be a Path object.")
        self.assertTrue(cache_dir.is_dir(), "The cache directory should exist.")
        self.assertEqual(expected_cache_name, cache_dir.name, "The cache directory should be named correctly.")
        self.assertEqual(self.CACHE_DIR, cache_dir.parent, "The cache directory should be in the correct location.")

    def test_clear_scoped_cache(self) -> None:
        """Test clearing the cache for a given scope."""
        scope = "some other scope"
        scope_2 = "another some other scope"
        cache_dir = self.cache.get_location(self.environment, scope)
        cache_dir_2 = self.cache.get_location(self.environment, scope_2)
        self.assertTrue(cache_dir.is_dir(), "The cache directory should exist.")
        self.assertTrue(cache_dir_2.is_dir(), "The cache directory should exist.")

        some_file = cache_dir / "some_file.txt"
        some_file.touch()
        self.assertTrue(some_file.exists(), "The file should exist.")

        self.cache.clear_cache(self.environment, scope)

        self.assertFalse(cache_dir.exists(), "The cache directory should be removed.")
        self.assertFalse(some_file.exists(), "The file should be removed.")
        self.assertTrue(self.CACHE_DIR.is_dir(), "The root cache should still exist.")
        self.assertTrue(cache_dir_2.is_dir(), "The other cache directory should still exist.")

    def test_clear_root_cache(self) -> None:
        """Test clearing the entire cache."""
        scope = "some other scope"
        cache_dir = self.cache.get_location(self.environment, scope)
        self.assertTrue(cache_dir.is_dir(), "The cache directory should exist.")

        some_file = cache_dir / "some_file.txt"
        some_file.touch()
        self.assertTrue(some_file.exists(), "The file should exist.")

        self.cache.clear_cache()

        self.assertFalse(cache_dir.exists(), "The cache directory should be removed.")
        self.assertFalse(some_file.exists(), "The file should be removed.")
        self.assertTrue(self.CACHE_DIR.is_dir(), "The root cache should still exist.")
        self.assertEqual(0, len(list(self.CACHE_DIR.iterdir())), "The cache should be empty.")

    def test_clear_cache_no_scope_raises_valueerror(self) -> None:
        """Test clearing the cache without specifying a scope should raise an error."""
        with self.assertRaises(ValueError):
            self.cache.clear_cache(environment=self.environment)

    def test_clear_cache_no_environment_raises_valueerror(self) -> None:
        """Test clearing the cache without specifying an environment should raise an error."""
        with self.assertRaises(ValueError):
            self.cache.clear_cache(scope="some scope")

    def test_temporary_cache(self) -> None:
        """Test creating a temporary cache."""
        with self.cache.temporary_location() as temp_cache:
            self.assertTrue(temp_cache.is_dir(), "The temporary cache directory should exist.")
            self.assertEqual(self.cache.root, temp_cache.parent, "The temporary cache should be in the root.")

        self.assertFalse(
            temp_cache.exists(), "The temporary cache directory should be removed after exiting the context."
        )
        self.assertTrue(self.CACHE_DIR.is_dir(), "The root cache should still exist.")



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from evo.common.io import Upload
from evo.common.test_tools import TestWithUploadHandler, UrlGenerator

from .common import TEST_DATA


class MyUpload(Upload):
    def __init__(self, url_generator: UrlGenerator) -> None:
        self._generator = url_generator

    @property
    def label(self) -> str:
        return "test_upload"

    async def get_upload_url(self) -> str:
        return await self._generator.get_new_url()


class TestUpload(TestWithUploadHandler):
    def setUp(self) -> None:
        super().setUp()
        self.upload = MyUpload(self.url_generator)

    async def test_from_path(self) -> None:
        test_data_file = self.CACHE_DIR / "test_data_upload.csv"
        test_data_file.write_bytes(TEST_DATA)
        await self.upload.upload_from_path(test_data_file, self.transport)
        uploaded_data = await self.handler.get_committed()
        self.assertEqual(test_data_file.read_bytes(), uploaded_data)



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from evo.blockmodels.endpoints.models import JobResponse, JobStatus
from evo.common.test_tools import AbstractTestRequestHandler, MockResponse


class JobPollingRequestHandler(AbstractTestRequestHandler):
    def __init__(self, job_response: JobResponse, pending_request: int = 0) -> None:
        self._job_response = job_response
        self._pending_requests = pending_request

    def job_poll(self) -> MockResponse:
        if self._pending_requests > 0:
            self._pending_requests -= 1
            job_response = JobResponse(job_status=JobStatus.PROCESSING)
        else:
            job_response = self._job_response
        return MockResponse(status_code=200, content=job_response.model_dump_json())



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import copy
import platform
import sys
import traceback
import unittest
from collections.abc import Callable, Sequence

from evo.common.exceptions import EvoExceptionGroup


class MyExc1(Exception): ...


class MyExc2(Exception): ...


class MyExc3(MyExc1): ...


MY_EXC_TYPES = (MyExc1, MyExc2, MyExc3)


def tb_text(exc: Exception) -> str:
    return "".join(traceback.format_exception(exc))


class TestEvoExceptionGroup(unittest.TestCase):
    EGROUP_TYPE = EvoExceptionGroup

    @classmethod
    def filter_excs(
        cls, excs: Sequence[Exception, ...], filter_func: Callable[[Exception], bool], inverse: bool = False
    ) -> tuple[Exception, ...]:
        flattened = []
        for exc in excs:
            if isinstance(exc, cls.EGROUP_TYPE):
                matched, unmatched = exc.split(filter_func)
                if matched and not inverse:
                    flattened.append(matched)
                    continue
                elif unmatched and inverse:
                    flattened.append(unmatched)
                    continue
            if filter_func(exc) ^ inverse:
                flattened.append(exc)
        return tuple(flattened)

    def raise_excs(self, excs: list[Exception]) -> EGROUP_TYPE:
        with self.assertRaises(self.EGROUP_TYPE) as cm:
            raise self.EGROUP_TYPE("Some exceptions occurred", excs)
        return cm.exception

    def compare_excs(self, expected: Exception, actual: Exception) -> None:
        if isinstance(expected, self.EGROUP_TYPE) or isinstance(actual, self.EGROUP_TYPE):
            self.assertEqual(expected.message, actual.message)
            self.assertEqual(expected.__traceback__, actual.__traceback__)
            self.assertEqual(expected.__context__, actual.__context__)
            self.assertEqual(expected.__cause__, actual.__cause__)
            self.assertEqual(
                len(expected.exceptions),
                len(actual.exceptions),
                f"Number of exceptions differ\n{tb_text(expected)} != {tb_text(actual)}",
            )
            for sub_expected, sub_actual in zip(expected.exceptions, actual.exceptions):
                self.compare_excs(sub_expected, sub_actual)
        else:
            self.assertEqual(expected, actual)

    def check_derived(self, parent: EGROUP_TYPE, derived: EGROUP_TYPE, expected_exceptions: tuple[Exception, ...]):
        self.assertIsInstance(derived.exceptions, tuple)

        expected = parent.derive(expected_exceptions)
        # Copy the original context, cause, and traceback.
        expected.__context__ = copy.copy(parent.__context__)
        expected.__cause__ = copy.copy(parent.__cause__)
        expected.__traceback__ = copy.copy(parent.__traceback__)

        self.compare_excs(expected, derived)

    def test_single_wrapped_exception(self) -> None:
        excs = [MyExc1("Single exception")]
        grp = self.raise_excs(excs)

        with self.subTest("field: message"):
            self.assertEqual("Some exceptions occurred", grp.message)

        with self.subTest("field: exceptions"):
            self.assertTupleEqual(tuple(excs), grp.exceptions)

        with self.subTest("method: subgroup"):
            self.check_derived(grp, grp.subgroup(MyExc1), tuple(excs))
            self.assertIsNone(grp.subgroup(MyExc2))
            self.check_derived(grp, grp.subgroup(self.EGROUP_TYPE), grp.exceptions)

        with self.subTest("method: split"):
            matched, unmatched = grp.split(MyExc1)
            self.check_derived(grp, matched, tuple(excs))
            self.assertIsNone(unmatched)

            matched, unmatched = grp.split(self.EGROUP_TYPE)
            self.check_derived(grp, matched, grp.exceptions)
            self.assertIsNone(unmatched)

    def test_multiple_same_exception(self) -> None:
        excs = [MyExc1(f"Multiple same exception {n}") for n in range(10)]
        grp = self.raise_excs(excs)

        with self.subTest("field: exceptions"):
            self.assertTupleEqual(tuple(excs), grp.exceptions)

        with self.subTest("method: subgroup"):
            self.check_derived(grp, grp.subgroup(MyExc1), tuple(excs))
            self.assertIsNone(grp.subgroup(MyExc2))
            self.check_derived(grp, grp.subgroup(self.EGROUP_TYPE), grp.exceptions)

        with self.subTest("method: split"):
            matched, unmatched = grp.split(MyExc1)
            self.check_derived(grp, matched, tuple(excs))
            self.assertIsNone(unmatched)

            matched, unmatched = grp.split(self.EGROUP_TYPE)
            self.check_derived(grp, matched, grp.exceptions)
            self.assertIsNone(unmatched)

    def test_multiple_different_exceptions(self) -> None:
        excs = []
        for n in range(10):
            msg = f"Multiple different exceptions {n}"
            exc_type = MY_EXC_TYPES[n % 2]
            excs.append(exc_type(msg))
        grp = self.raise_excs(excs)

        with self.subTest("field: exceptions"):
            self.assertTupleEqual(tuple(excs), grp.exceptions)

        tuple_myexc1 = self.filter_excs(excs, lambda e: isinstance(e, MyExc1))
        tuple_myexc2 = self.filter_excs(excs, lambda e: isinstance(e, MyExc2))
        tuple_msg = self.filter_excs(excs, lambda e: str(e) == msg)
        tuple_not_msg = self.filter_excs(excs, lambda e: str(e) != msg)

        with self.subTest("method: subgroup"):
            self.check_derived(grp, grp.subgroup(MyExc1), tuple_myexc1)
            self.check_derived(grp, grp.subgroup(MyExc2), tuple_myexc2)
            self.check_derived(grp, grp.subgroup((MyExc1, MyExc2)), tuple(excs))
            self.check_derived(grp, grp.subgroup(lambda e: str(e) == msg), tuple_msg)
            self.check_derived(grp, grp.subgroup(self.EGROUP_TYPE), grp.exceptions)

        with self.subTest("method: split"):
            matched, unmatched = grp.split(MyExc1)
            self.check_derived(grp, matched, tuple_myexc1)
            self.check_derived(grp, unmatched, tuple_myexc2)

            matched, unmatched = grp.split(MyExc2)
            self.check_derived(grp, matched, tuple_myexc2)
            self.check_derived(grp, unmatched, tuple_myexc1)

            matched, unmatched = grp.split((MyExc1, MyExc2))
            self.check_derived(grp, matched, tuple(excs))
            self.assertIsNone(unmatched)

            matched, unmatched = grp.split(lambda e: str(e) == msg)
            self.check_derived(grp, matched, tuple_msg)
            self.check_derived(grp, unmatched, tuple_not_msg)

            matched, unmatched = grp.split(self.EGROUP_TYPE)
            self.check_derived(grp, matched, grp.exceptions)
            self.assertIsNone(unmatched)

    def test_multiple_different_exceptions_incl_subclass(self) -> None:
        excs = []
        for n in range(10):
            msg = f"Multiple different exceptions including subclasses {n}"
            exc_type = MY_EXC_TYPES[n % 2]
            excs.append(exc_type(msg))
        grp = self.raise_excs(excs)

        with self.subTest("field: exceptions"):
            self.assertTupleEqual(tuple(excs), grp.exceptions)

        tuple_myexc1_myexc3 = self.filter_excs(excs, lambda e: isinstance(e, (MyExc1, MyExc3)))
        tuple_myexc2 = self.filter_excs(excs, lambda e: isinstance(e, MyExc2))
        tuple_msg = self.filter_excs(excs, lambda e: str(e) == msg)
        tuple_not_msg = self.filter_excs(excs, lambda e: str(e) == msg, inverse=True)

        with self.subTest("method: subgroup"):
            self.check_derived(grp, grp.subgroup(MyExc1), tuple_myexc1_myexc3)
            self.check_derived(grp, grp.subgroup(MyExc2), tuple_myexc2)
            self.check_derived(grp, grp.subgroup((MyExc1, MyExc2)), tuple(excs))
            self.check_derived(grp, grp.subgroup(lambda e: str(e) == msg), tuple_msg)
            self.check_derived(grp, grp.subgroup(self.EGROUP_TYPE), grp.exceptions)

        with self.subTest("method: split"):
            matched, unmatched = grp.split(MyExc1)
            self.check_derived(grp, matched, tuple_myexc1_myexc3)
            self.check_derived(grp, unmatched, tuple_myexc2)

            matched, unmatched = grp.split(MyExc2)
            self.check_derived(grp, matched, tuple_myexc2)
            self.check_derived(grp, unmatched, tuple_myexc1_myexc3)

            matched, unmatched = grp.split((MyExc1, MyExc2))
            self.check_derived(grp, matched, tuple(excs))
            self.assertIsNone(unmatched)

            matched, unmatched = grp.split(lambda e: str(e) == msg)
            self.check_derived(grp, matched, tuple_msg)
            self.check_derived(grp, unmatched, tuple_not_msg)

            matched, unmatched = grp.split(self.EGROUP_TYPE)
            self.check_derived(grp, matched, grp.exceptions)
            self.assertIsNone(unmatched)

    def test_subgroups(self) -> None:
        excs = []
        for i in range(10):
            subgrp_excs = []
            for j in range(i + 1):
                exc_msg = f"Exc {i + 1}.{j + 1}"
                exc_type = MY_EXC_TYPES[(i * j) % 3]
                subgrp_excs.append(exc_type(exc_msg))
            excs.append(self.EGROUP_TYPE(f"Subgroup {i + 1}", subgrp_excs))

        grp = self.raise_excs(excs)

        with self.subTest("field: exceptions"):
            self.assertTupleEqual(tuple(excs), grp.exceptions)

        tuple_myexc1_myexc3 = self.filter_excs(excs, lambda e: isinstance(e, (MyExc1, MyExc3)))
        tuple_myexc2 = self.filter_excs(excs, lambda e: isinstance(e, MyExc2))
        tuple_msg = self.filter_excs(excs, lambda e: str(e) == exc_msg)
        tuple_not_msg = self.filter_excs(excs, lambda e: str(e) == exc_msg, inverse=True)

        with self.subTest("method: subgroup"):
            self.check_derived(grp, grp.subgroup(MyExc1), tuple_myexc1_myexc3)
            self.check_derived(grp, grp.subgroup(MyExc2), tuple_myexc2)
            self.check_derived(grp, grp.subgroup((MyExc1, MyExc2)), tuple(excs))
            self.check_derived(grp, grp.subgroup(lambda e: str(e) == exc_msg), tuple_msg)
            self.check_derived(grp, grp.subgroup(self.EGROUP_TYPE), grp.exceptions)

        with self.subTest("method: split"):
            matched, unmatched = grp.split(MyExc1)
            self.check_derived(grp, matched, tuple_myexc1_myexc3)
            self.check_derived(grp, unmatched, tuple_myexc2)

            matched, unmatched = grp.split(MyExc2)
            self.check_derived(grp, matched, tuple_myexc2)
            self.check_derived(grp, unmatched, tuple_myexc1_myexc3)

            matched, unmatched = grp.split((MyExc1, MyExc2))
            self.check_derived(grp, matched, tuple(excs))
            self.assertIsNone(unmatched)

            matched, unmatched = grp.split(lambda e: str(e) == exc_msg)
            self.check_derived(grp, matched, tuple_msg)
            self.check_derived(grp, unmatched, tuple_not_msg)

            matched, unmatched = grp.split(self.EGROUP_TYPE)
            self.check_derived(grp, matched, grp.exceptions)
            self.assertIsNone(unmatched)

    def test_nested_subgroups(self) -> None:
        excs = []
        for i in range(10):
            subgrp_i_excs = []
            for j in range(i + 1):
                subgrp_j_excs = []
                for k in range(j + 1):
                    exc_msg = f"Exc {i + 1}.{j + 1}.{k + 1}"
                    exc_type = MY_EXC_TYPES[(i * j) % 3]
                    subgrp_j_excs.append(exc_type(exc_msg))
                subgrp_i_excs.append(self.EGROUP_TYPE(f"Subgroup {i + 1}.{j + 1}", subgrp_j_excs))
            excs.append(self.EGROUP_TYPE(f"Subgroup {i + 1}", subgrp_i_excs))

        grp = self.raise_excs(excs)

        with self.subTest("field: exceptions"):
            self.assertTupleEqual(tuple(excs), grp.exceptions)

        tuple_myexc1_myexc3 = self.filter_excs(excs, lambda e: isinstance(e, (MyExc1, MyExc3)))
        tuple_myexc2 = self.filter_excs(excs, lambda e: isinstance(e, MyExc2))
        tuple_msg = self.filter_excs(excs, lambda e: str(e) == exc_msg)
        tuple_not_msg = self.filter_excs(excs, lambda e: str(e) == exc_msg, inverse=True)

        with self.subTest("method: subgroup"):
            self.check_derived(grp, grp.subgroup(MyExc1), tuple_myexc1_myexc3)
            self.check_derived(grp, grp.subgroup(MyExc2), tuple_myexc2)
            self.check_derived(grp, grp.subgroup((MyExc1, MyExc2)), tuple(excs))
            self.check_derived(grp, grp.subgroup(lambda e: str(e) == exc_msg), tuple_msg)
            self.check_derived(grp, grp.subgroup(self.EGROUP_TYPE), grp.exceptions)

        with self.subTest("method: split"):
            matched, unmatched = grp.split(MyExc1)
            self.check_derived(grp, matched, tuple_myexc1_myexc3)
            self.check_derived(grp, unmatched, tuple_myexc2)

            matched, unmatched = grp.split(MyExc2)
            self.check_derived(grp, matched, tuple_myexc2)
            self.check_derived(grp, unmatched, tuple_myexc1_myexc3)

            matched, unmatched = grp.split((MyExc1, MyExc2))
            self.check_derived(grp, matched, tuple(excs))
            self.assertIsNone(unmatched)

            matched, unmatched = grp.split(lambda e: str(e) == exc_msg)
            self.check_derived(grp, matched, tuple_msg)
            self.check_derived(grp, unmatched, tuple_not_msg)

            matched, unmatched = grp.split(self.EGROUP_TYPE)
            self.check_derived(grp, matched, grp.exceptions)
            self.assertIsNone(unmatched)


@unittest.skipUnless(
    sys.version_info >= (3, 11), f"ExceptionGroup is not implemented in python {platform.python_version()}"
)
class TestBuiltinExceptionGroup(TestEvoExceptionGroup):
    """Run the same tests on the builtin ExceptionGroup type when it is available"""

    @classmethod
    def setUpClass(cls) -> None:
        # Defer this assignment until runtime to prevent an exception in earlier versions of python.
        cls.EGROUP_TYPE = ExceptionGroup  # noqa: F821
